
from django.test import TestCase
from django.http import HttpResponse
import pytest
from unittest.mock import Mock, call

from lily.base.events import EventFactory


class GenericExceptionTestCase(TestCase):

    def test_init(self):

        e = EventFactory.Generic(status_code=401, content='{"hello":"there"}')

        assert e.status_code == 401
        assert e.content == '{"hello":"there"}'

    def test_extend(self):

        e = EventFactory.Generic(status_code=401, content='{"hello":"there"}')
        extended = e.extend(method='POST', path='/hi/there')

        assert extended.method == 'POST'
        assert extended.path == '/hi/there'
        assert isinstance(extended, EventFactory.Generic)

    def test_log(self):

        e = EventFactory.Generic(status_code=401, content='{"hello":"there"}')
        e.logger = Mock()

        e.extend(method='POST', path='/hi/there').log()

        assert e.logger.info.call_args_list == (
            [call('[POST /hi/there] -> 401')])

    def test_response(self):
        e = EventFactory.Generic(status_code=401, content='{"hello":"there"}')

        response = e.response()

        assert isinstance(response, HttpResponse)
        assert response.status_code == 401
        assert response.content == b'{"hello":"there"}'


class ContextTestCase(TestCase):

    def test_constructor(self):

        c = EventFactory.Context(user_id=12, email='a@p.a', origin='here')

        assert c.data['user_id'] == 12
        assert c.data['email'] == 'a@p.a'
        assert c.data['origin'] == 'here'

    def test_is_empty(self):

        assert EventFactory.Context().is_empty() is True
        assert EventFactory.Context(user_id=11).is_empty() is False
        assert EventFactory.Context(origin='here').is_empty() is False
        assert EventFactory.Context(email='here@here').is_empty() is False


class BaseSuccessExceptionTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker):
        self.mocker = mocker

    def test_create_empty__defaults(self):

        e = EventFactory.BaseSuccessException()

        assert e.data == {}
        assert e.context == EventFactory.Context()
        assert e.event is None
        assert e.instance is None

    def test_extend(self):

        context = Mock()
        e = EventFactory.BaseSuccessException()
        prev_context = e.context

        e.extend(event='HI', context=context)

        assert e.event == 'HI'
        assert e.context == context
        assert e.context != prev_context

    def test_log__user_id_in_context(self):

        dumps = self.mocker.patch('lily.base.events.json.dumps')
        dumps.return_value = '{XX}'
        e = EventFactory.BaseSuccessException(
            context=Mock(log_authorizer={'user_id': 12}), event='HELLO')
        logger = self.mocker.patch.object(e, 'logger')

        e.log()

        assert logger.info.call_args_list == [call('HELLO: {XX}')]
        assert dumps.call_args_list == [
            call({'@authorizer': {'user_id': 12}, '@event': 'HELLO'})]

    def test_log__no_user_id_in_context(self):

        dumps = self.mocker.patch('lily.base.events.json.dumps')
        dumps.return_value = '{XX}'
        logger = Mock()
        e = EventFactory.BaseSuccessException(
            context=EventFactory.Context(), event='HELLO')
        logger = self.mocker.patch.object(e, 'logger')

        e.log()

        assert logger.info.call_args_list == [call('HELLO: {XX}')]
        assert dumps.call_args_list == [call({'@event': 'HELLO'})]


@pytest.mark.parametrize(
    'exception, expected_status_code', [

        #
        # SUCCESS RESPONSES
        #
        (EventFactory.Executed, 200),

        (EventFactory.Created, 201),
        (EventFactory.Read, 200),
        (EventFactory.Updated, 200),
        (EventFactory.Deleted, 200),

        (EventFactory.BulkCreated, 201),
        (EventFactory.BulkRead, 200),
        (EventFactory.BulkUpdated, 200),
        (EventFactory.BulkDeleted, 200),

        #
        # ERROR RESPONSES
        #
        (EventFactory.BrokenRequest, 400),
        (EventFactory.AuthError, 401),
        (EventFactory.AccessDenied, 403),
        (EventFactory.DoesNotExist, 404),
        (EventFactory.Conflict, 409),
        (EventFactory.ServerError, 500),
    ])
def test_response_classes(exception, expected_status_code):

    assert exception.response_class.status_code == expected_status_code




import os
import re

from .utils import to_camelcase


class Path:

    def __init__(self, base_path):
        self.base_path = base_path

    def join(self, sub_path):
        sub_path = re.sub('^/', '', sub_path)
        return os.path.join(self.base_path, sub_path)


class Domain:

    def __init__(self, domain_id, domain_name):
        self.id = domain_id
        self.name = domain_name
        self.camel_id = to_camelcase(domain_id)

    @property
    def path(self):
        return Path(
            './projects/client/src/domains/{}'.format(self.id))

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id




import math
from enum import EnumMeta

from django.db import models
from django.db.models.expressions import RawSQL
from django.contrib.postgres.fields import JSONField
from jsonschema import (
    validate as json_validate,
    ValidationError as JsonValidationError,
)
from django.core.exceptions import ValidationError


class ImmutableModel(models.Model):

    class ModelIsImmutableError(Exception):
        pass

    def save(self, *args, **kwargs):
        self.full_clean()

        if self.id is not None:
            raise ImmutableModel.ModelIsImmutableError(
                '{} model is declared as immutable'.format(
                    self.__class__.__name__))

        super(ImmutableModel, self).save(*args, **kwargs)

    class Meta:
        abstract = True


class ValidatingModel(models.Model):

    def save(self, *args, **kwargs):

        # -- call all clean & validation methods without capturing
        # -- their individual ValidationErrors
        try:
            self.clean_fields()

        except ValidationError as e:
            raise ValidationError(e.update_error_dict({}))

        try:
            self.clean()

        except ValidationError as e:
            raise ValidationError(e.update_error_dict({}))

        try:
            self.validate_unique()

        except ValidationError as e:
            raise ValidationError(e.update_error_dict({}))

        return super(ValidatingModel, self).save(*args, **kwargs)

    class Meta:
        abstract = True


class ExtraColumn(RawSQL):

    def __init__(self, sql, params, output_field=None):

        super(ExtraColumn, self).__init__(
            sql, params, output_field=output_field)

    def get_group_by_cols(self):
        return []


class JSONSchemaValidator:

    validation_error_cls = ValidationError

    def __init__(self, schema):
        self.schema = schema

    def __call__(self, value):

        # -- multischema case
        if 'schemas' in self.schema:
            by_field = self.schema['by_field']
            schemas = self.schema['schemas']
            by_field_values = sorted(schemas.keys())

            self.validate_against_schema(
                object(**{by_field: enum(*by_field_values)}),
                {by_field: value[by_field]})

            for schema_key, schema in schemas.items():
                if value[by_field] == schema_key:
                    self.validate_against_schema(schema, value)

        # -- array multischema
        elif (self.schema.get('type') == 'array' and
                self.schema.get('items', {}).get('oneOf') and
                self.schema.get('items', {}).get('by_field')):

            # -- for now only support for the enum fields
            by_field = self.schema['items'].get('by_field')

            by_field_values = sorted([
                schema['properties'][by_field].get('const_value')
                for schema in self.schema['items']['oneOf']
            ])
            schemas_index = {
                schema['properties'][by_field].get('const_value'): schema
                for schema in self.schema['items']['oneOf']
            }

            # -- assert its an array
            self.validate_against_schema(array(object()), value)
            for item in value:
                fields = {by_field: enum(*by_field_values)}

                # -- assert required
                self.validate_against_schema(
                    object(**fields, required=[by_field]), item)

                # -- assert in enum
                self.validate_against_schema(
                    object(**fields), {by_field: item[by_field]})

                schema = schemas_index[item[by_field]]
                self.validate_against_schema(schema, item)

        # -- single schema case
        else:
            self.validate_against_schema(self.schema, value)

    def validate_against_schema(self, schema, value):
        try:
            json_validate(value, schema)

        except JsonValidationError as e:
            path = '.'.join([str(p) for p in e.path]) or '.'
            raise self.validation_error_cls(
                f"JSON did not validate. PATH: '{path}' REASON: {e.message}"
            )


class JSONSchemaField(JSONField):
    def __init__(self, *args, **kwargs):
        try:
            schema = kwargs.pop('schema')

        except KeyError:
            schema = {}

        super(JSONSchemaField, self).__init__(*args, **kwargs)
        self.validators.insert(0, JSONSchemaValidator(schema=schema))


class EnumChoiceField(models.CharField):

    def __init__(self, *args, enum_name=None, enum=None, **kwargs):

        self.enum_name = None
        if enum_name:
            self.enum_name = enum_name

        if enum:
            self.enum_name = enum.__name__
            kwargs['choices'] = [(e.value, e.value) for e in enum]

        # -- setting dynamical max_length
        if not kwargs.get('max_length'):
            max_length = 0
            for choice in kwargs['choices']:
                if len(str(choice[0])) > max_length:
                    max_length = len(str(choice[0]))

            kwargs['max_length'] = 2 ** math.ceil(math.log2(max_length))

        super(EnumChoiceField, self).__init__(*args, **kwargs)

    def deconstruct(self):

        name, path, args, kwargs = super().deconstruct()
        kwargs['enum_name'] = self.enum_name

        return name, path, args, kwargs


#
# Helper class for building easier to read JSON schema
#
def null_or(other):
    return {
        'oneOf': [
            {
                'type': 'null',
            },
            other,
        ],
    }


def null():
    return {
        'type': 'null',
    }


def string(
        pattern=None,
        minLength=None,  # noqa
        maxLength=None):  # noqa
    schema = {
        'type': 'string',
    }

    if pattern is not None:
        schema['pattern'] = pattern

    if minLength is not None:
        schema['minLength'] = minLength

    if maxLength is not None:
        schema['maxLength'] = maxLength

    return schema


def const(value):
    return {
        'type': 'string',
        'const': value,
        'pattern': f'^{value}$'
    }


def number(
        multipleOf=None,  # noqa
        minimum=None,
        exclusiveMinimum=None,  # noqa
        maximum=None,
        exclusiveMaximum=None):  # noqa

    schema = {
        'type': 'number',
    }

    if multipleOf is not None:
        schema['multipleOf'] = multipleOf

    if minimum is not None:
        schema['minimum'] = minimum

    if exclusiveMinimum is not None:
        schema['exclusiveMinimum'] = exclusiveMinimum

    if maximum is not None:
        schema['maximum'] = maximum

    if exclusiveMaximum is not None:
        schema['exclusiveMaximum'] = exclusiveMaximum

    return schema


def boolean():
    return {
        'type': 'boolean',
    }


def url():
    return {
        'type': 'string',
        'format': 'url',
    }


def enum(*enums, enum_name=None, const=None):

    if len(enums) == 1 and isinstance(enums[0], EnumMeta):
        enum = enums[0]
        enum_name = enum.__name__
        enums = [e.value for e in enum]

    if isinstance(enums[0], int):
        _type = 'integer'

    else:
        _type = 'string'

    output = {
        'type': _type,
        'enum_name': enum_name,
        'enum': list(enums),
    }

    if const:
        output['const_value'] = const

    return output


def object(dependencies=None, required=None, **properties):
    schema = {
        'type': 'object',
        'properties': properties,
    }

    if required:
        schema['required'] = required

    if dependencies:
        schema['dependencies'] = dependencies

    return schema


def dependencies(**deps):
    return deps


def array(items, extra=None):
    extra = extra or {}

    return {
        'type': 'array',
        'items': items,
        **extra,
    }


def multischema(schemas, by_field):
    return {
        'schemas': schemas,
        'by_field': by_field,
        'oneOf': list(schemas.values()),
    }


def one_of(*schemas, by_field=None):
    return {
        'oneOf': list(schemas),
        'by_field': by_field,
    }




import json
import os

import djclick as click
from lily_assistant.config import Config

from ...renderer import CommandsRenderer
from ...serializers import CommandSerializer


@click.command()
def command():
    """Render all commands spec.

    Render all commands with their spec, examples, schemas etc. so far created
    for a current version of the service.

    """

    config = Config()

    commands_dir_path = os.path.join(Config.get_lily_path(), 'commands')
    if not os.path.exists(commands_dir_path):
        os.mkdir(commands_dir_path)

    version = config.next_version or config.version
    commands_path = os.path.join(commands_dir_path, f'{version}.json')
    with open(commands_path, 'w') as f:
        commands = CommandsRenderer().render()
        enums = commands.pop('@enums')

        commands = {
            name: CommandSerializer(conf).data
            for name, conf in commands.items()
        }
        f.write(
            json.dumps(
                {
                    **commands,
                    '@enums': enums,
                },
                indent=4,
                sort_keys=False))

    click.secho(
        f'Commands rendered for to file {commands_path}',
        fg='green')




import os
import json

from lily_assistant.config import Config

from lily.conf import settings
from lily.base import serializers, parsers, name
from lily.base.command import command
from lily.base.commands import HTTPCommands
from lily.base.meta import Meta, Domain
from lily.base.access import Access
from lily.base.input import Input
from lily.base.output import Output
from .serializers import CommandSerializer


class EntryPointCommands(HTTPCommands):

    class EntryPointSerializer(serializers.Serializer):

        class VersionInfoSerializer(serializers.Serializer):

            _type = 'version_info'

            deployed = serializers.CharField()

            displayed = serializers.CharField()

            available = serializers.ListField(child=serializers.CharField())

        _type = 'entrypoint'

        version_info = VersionInfoSerializer()

        name = serializers.CharField()

        commands = serializers.DictField(child=CommandSerializer())

        at__enums = serializers.ListField(child=serializers.DictField())

    class QueryParser(parsers.Parser):

        commands = parsers.ListField(child=parsers.CharField(), default=None)

        with_schemas = parsers.BooleanField(default=True)

        is_private = parsers.BooleanField(default=None)

        domain_id = parsers.CharField(default=None)

        version = parsers.CharField(default=None)

    @command(
        name=name.Read('ENTRY_POINT'),

        meta=Meta(
            title='Read Entry Point',
            description='''
                Serve Service Entry Point data:
                - current or chosen version of the service
                - list of all available commands together with their
                  configurations
                - examples collected for a given service.

            ''',
            domain=Domain(id='docs', name='Docs Management')),

        access=Access(
            is_private=True,
            access_list=settings.LILY_ENTRYPOINT_COMMANDS_ACCESS_LIST),

        input=Input(query_parser=QueryParser),

        output=Output(serializer=EntryPointSerializer),
    )
    def get(self, request):

        command_names = request.input.query['commands']

        is_private = request.input.query['is_private']

        domain_id = request.input.query['domain_id']

        version = request.input.query['version']

        config = Config()

        commands = self.get_commands(version)
        enums = commands.pop('@enums')

        if command_names:
            commands = {
                command_name: commands[command_name]
                for command_name in command_names}

        if is_private is not None:
            commands = {
                name: command
                for name, command in commands.items()
                if command['access']['is_private'] == is_private
            }

        if domain_id:
            commands = {
                name: command
                for name, command in commands.items()
                if command['meta']['domain']['id'].lower() == domain_id.lower()
            }

        raise self.event.Read(
            {
                'name': config.name,
                'version_info': {
                    'deployed': config.version,
                    'displayed': version or config.version,
                    'available': self.get_available_versions(),
                },
                'commands': commands,
                'at__enums': enums,
            })

    def get_available_versions(self):

        commands_dir_path = os.path.join(Config.get_lily_path(), 'commands')

        return sorted(
            [
                commands_file.replace('.json', '')
                for commands_file in os.listdir(commands_dir_path)
            ],
            key=lambda x: [int(e) for e in x.split('.')],
            reverse=True)

    def get_commands(self, version=None):

        config = Config()
        version = version or config.version
        commands_dir_path = os.path.join(Config.get_lily_path(), 'commands')

        commands_path = os.path.join(commands_dir_path, f'{version}.json')
        with open(commands_path, 'r') as f:
            return json.loads(f.read())




import inspect
from uuid import uuid4

from django.http import HttpRequest


# FIXME: test it!!!
class Context:

    def __init__(self, command_name, request):

        # -- to track current command
        self.command_name = command_name

        # -- to track requests send between commands
        if request.META.get('HTTP_X_CS_CORRELATION_ID'):
            self.correlation_id = request.META['HTTP_X_CS_CORRELATION_ID']

        else:
            self.correlation_id = str(uuid4())


# FIXME: test it!!!
def get_context():
    frame = inspect.currentframe()

    while frame:
        request = frame.f_locals.get('request')

        if isinstance(request, HttpRequest):
            return request._lily_context




from unittest.mock import Mock

from django.test import TestCase
import pytest

from lily.base.meta import Meta, Domain
from lily.docs.renderers.markdown.renderer import MarkdownRenderer
from lily.entrypoint.base import BaseRenderer
from tests import remove_white_chars


class MarkdownRendererTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def test_render(self):

        self.mocker.patch.object(BaseRenderer, 'render').return_value = {
            'LIST_ITEMS': {
                'path_conf': {
                    'path': '/items/',
                    'parameters': {}
                },
                'method': 'get',
                'meta': Meta(
                    title='hi there',
                    description='what?',
                    domain=Domain(id='d', name='items management')),
            },
            'CREATED': {
                'path_conf': {
                    'path': '/items/',
                    'parameters': {}
                },
                'method': 'post',
                'meta': Meta(
                    title='hi there',
                    description='what?',
                    domain=Domain(id='d', name='items management')),
            },
        }
        self.mocker.patch.object(
            MarkdownRenderer, 'get_examples'
        ).return_value = {
            'CREATED': {
                '201 (CREATED)': {
                    'response': {
                        'content_type': 'application/json',
                        'status': 201,
                        'content': {
                            'user_id': 434,
                            '@event': 'CREATED',
                        }
                    },
                    'request': {
                        'path': 'hello/hi',
                        'headers': {
                            'HI-THERE': 'JSON',
                        },
                        'content': {
                            'hi': 'there'
                        }
                    },
                    'description': 'CREATE',
                    'method': 'post'
                },
            },
            'LIST_ITEMS': {
                '200 (LISTED)': {
                    'response': {
                        'content_type': 'application/json',
                        'status': 200,
                        'content': {
                            'user_id': 434,
                            '@event': 'LISTED',
                        }
                    },
                    'request': {
                        'path': 'wat/178',
                        'headers': {
                            'HI-THERE': 'JSON',
                        }
                    },
                    'description': 'SERVER_ERROR',
                    'method': 'get'
                },
                '502 (SERVER_ERROR)': {
                    'response': {
                        'content_type': 'application/json',
                        'status': 502,
                        'content': {
                            'user_id': 434,
                            '@type': 'error',
                            '@event': 'SERVER_ERROR',
                        }
                    },
                    'request': {
                        'path': 'hello/world',
                        'headers': {
                            'HI-THERE': 'JSON',
                        }
                    },
                    'description': 'SERVER_ERROR',
                    'method': 'get'
                },
            }
        }

        assert remove_white_chars(
            MarkdownRenderer(Mock()).render()
        ) == remove_white_chars('''

            # CoSphere API

            CoSphere's API with hypermedia links

            ## items management

            ### CREATED: POST /items/

            hi there
            what?

            #### 201 (CREATED)

            Request:
            ```http
            POST hello/hi HTTP/1.1
            HI-THERE: JSON
            {
                "hi": "there"
            }
            ```

            Respone:
            ```json
            {
                "@event": "CREATED",
                "user_id": 434
            }
            ```

            ### LIST_ITEMS: GET /items/
            hi there
            what?

            #### 200 (LISTED)

            Request:
            ```http
            GET wat/178 HTTP/1.1
            HI-THERE: JSON
            ```

            Respone:
            ```json
            {
                "@event": "LISTED",
                "user_id": 434
            }
            ```

            #### 502 (SERVER_ERROR)

            Request:
            ```http
            GET hello/world HTTP/1.1
            HI-THERE: JSON
            ```

            Respone:
            ```json
            {
                "@event": "SERVER_ERROR",
                "@type": "error",
                "user_id": 434
            }
            ```

        ''')




import json
from contextlib import ContextDecorator
from unittest.mock import Mock

from django.test import TestCase
from django.http import HttpResponse
from django.core.exceptions import ValidationError
from django.db.utils import DatabaseError
from django.contrib.auth.models import User
from django_fake_model import models as fake_models
from django.db import models, transaction
import pytest

from lily.base.command import command, HTTPCommands
from lily.base.access import Access
from lily.base.meta import Meta, Domain
from lily.base.input import Input
from lily.base.output import Output
from lily.base import serializers, parsers
from lily.base.events import EventFactory


event = EventFactory()


def to_json(response, content_field='content'):
    return json.loads(str(getattr(response, content_field), encoding='utf8'))


class Request:
    pass


def dump_to_bytes(data):
    return bytes(json.dumps(data), 'utf8')


def get_auth_headers(user_id, account_type='SUPER_PREMIUM'):
    return {
        'HTTP_X_CS_ACCOUNT_TYPE': account_type,
        'HTTP_X_CS_USER_ID': user_id,
        'SERVER_NAME': 'yo',
    }


class FakeClient(fake_models.FakeModel):
    name = models.CharField(max_length=100)


class GenericCommands(HTTPCommands):

    @command(
        name='GET_IT',
        meta=Meta(
            title='get',
            description='get it...',
            domain=Domain(id='get', name='get')),
    )
    def get(self, request):

        status_code = int(request.GET['status_code'])

        raise self.event.Generic(
            status_code=status_code,
            content=json.dumps({
                '@event': 'FOUND_IT',
                '@type': 'what',
                'hi': status_code,
            }))


class HttpCommands(HTTPCommands):

    @command(
        name='HTTP_IT',
        meta=Meta(
            title='http',
            description='http it...',
            domain=Domain(id='http', name='http')),
    )
    def get(self, request):

        return HttpResponse('hello world')


@FakeClient.fake_me
class TestCommands(HTTPCommands):

    class BodyParser(parsers.Parser):

        name = parsers.CharField()

        age = parsers.IntegerField()

    class ClientSerializer(serializers.ModelSerializer):

        _type = 'client'

        card_id = serializers.SerializerMethodField()

        def get_access(self, instance):
            return [(TestCommands.put, True)]

        class Meta:
            model = FakeClient
            fields = ('name', 'card_id')

        def get_card_id(self, instance):
            return 190

    class SimpleSerializer(serializers.Serializer):
        _type = 'simple'

        amount = serializers.IntegerField()

    @command(
        name='MAKE_IT',
        meta=Meta(
            title='hi there',
            description='it is made for all',
            domain=Domain(id='test', name='test management')),
        input=Input(body_parser=BodyParser),
        output=Output(serializer=ClientSerializer),
        access=Access(
            access_list=['PREMIUM', 'SUPER_PREMIUM'])
    )
    def post(self, request, user_id):
        raise self.event.Executed(
            event='MADE_IT',
            instance=FakeClient(name="Jake"))

    @command(
        name='GET_IT',
        meta=Meta(
            title='get',
            description='get it...',
            domain=Domain(id='get', name='get')),
        output=Output(serializer=ClientSerializer))
    def get(self, request):
        raise self.event.Executed(
            event='GET_IT',
            instance=FakeClient(name="Jake"))

    @command(
        name='BREAK',
        meta=Meta(
            title='break',
            description='break it...',
            domain=Domain(id='break', name='break')),
        output=Output(serializer=SimpleSerializer))
    def put(self, request):

        raise self.event.Executed(
            event='BROKEN',
            context=request,
            data=json.loads(request.body.decode('utf8')))

    @command(
        name='ATOMIC',
        meta=Meta(
            title='atomic',
            description='atomic it...',
            domain=Domain(id='atomic', name='atomic')),
        is_atomic='default')
    def delete(self, request):

        self.some_stuff()

        raise self.event.Executed(event='ATOMIC', context=request, data={})

    def some_stuff(self):
        pass


class DoAndGoCommands(HTTPCommands):

    @command(
        name='DO_AND_GO',
        meta=Meta(
            title='do and go',
            domain=Domain(id='test', name='test management')),
        access=Access(access_list=['PREMIUM'])
    )
    def post(self, request):

        raise self.event.Redirect(
            'DONE_SO_GO',
            redirect_uri='http://go.there.org')


@FakeClient.fake_me
class BrokenSerializerCommands(HTTPCommands):

    class ClientSerializer(serializers.Serializer):

        _type = 'client'

        card_id = serializers.SerializerMethodField()

        def get_card_id(self, instance):
            raise EventFactory.AccessDenied('GO_AWAY')

    @command(
        name='BREAK_SERIALIZER',
        meta=Meta(
            title='break it',
            domain=Domain(id='test', name='test management')),
        output=Output(serializer=ClientSerializer),
        access=Access(access_list=['PREMIUM', 'SUPER_PREMIUM'])
    )
    def post(self, request):
        raise self.event.Executed(
            event='NEVER_REACH_IT',
            instance=FakeClient(name="Jake"))


class CommandTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker):
        self.mocker = mocker

    def setUp(self):
        User.objects.all().delete()

    def test_success(self):

        u = User.objects.create_user(username='jacky')
        request = Mock(
            body=dump_to_bytes({"name": "John", "age": 81}),
            META=get_auth_headers(u.id, 'SUPER_PREMIUM'))
        self.mocker.patch.object(
            serializers,
            'COMMANDS_CONF',
            {
                'MAKE_IT_BETTER': {
                    'service_base_uri': 'http://192.11.2.1:9000',
                    'method': 'post',
                    'path_conf': {
                        'path': '/payment_cards/{card_id}/sth/',
                        'parameters': [
                            {
                                'name': 'card_id',
                                'in': 'path',
                                'description': None,
                                'required': True,
                                'type': 'integer'
                            },
                        ],
                    },
                    'access': {
                        'is_private': False,
                        'access_list': ['SUPER_PREMIUM'],
                    },
                },
            })

        c = TestCommands()

        response = c.post(request, 11)

        assert response.status_code == 200
        assert to_json(response) == {
            '@type': 'client',
            '@event': 'MADE_IT',
            '@access': {
                'BREAK': True,
            },
            'name': 'Jake',
            'card_id': 190,
        }

    #
    # REDIRECT
    #
    def test_redirect(self):

        request = Mock(META=get_auth_headers(11, 'PREMIUM'))

        response = DoAndGoCommands().post(request)

        assert response.status_code == 303
        assert to_json(response) == {
            '@authorizer': {'account_type': 'PREMIUM', 'user_id': 11},
            '@event': 'DONE_SO_GO',
            '@type': 'error',
        }

    #
    # BROKEN SERIALIZER
    #
    def test_exception_from_serializer(self):

        request = Mock(META=get_auth_headers(11, 'PREMIUM'))

        response = BrokenSerializerCommands().post(request)

        assert response.status_code == 403
        assert to_json(response) == {
            '@authorizer': {'account_type': 'PREMIUM', 'user_id': 11},
            '@event': 'GO_AWAY',
            '@type': 'error',
        }

    #
    # AUTHORIZATION
    #
    def test_access_denied__missing_headers(self):

        request = Request()
        request.body = dump_to_bytes({"name": "John", "age": 81})
        request.META = {}
        c = TestCommands()

        response = c.post(request, 12)

        assert response.status_code == 403
        assert to_json(response) == {
            '@type': 'error',
            '@event': 'ACCESS_DENIED',
        }

    def test_no_authorization(self):

        request = Request()
        request.body = dump_to_bytes({"name": "John", "age": 81})
        request.META = {}
        c = TestCommands()

        response = c.get(request)

        assert response.status_code == 200

    #
    # CONF
    #
    def test_conf__context_attached_to_the_request(self):

        u = User.objects.create_user(username='jacky')
        request = Request()
        request.body = dump_to_bytes({"name": "John", "age": 81})
        request.META = get_auth_headers(u.id, 'SUPER_PREMIUM')
        c = TestCommands()

        c.post(request, 15)

        assert request._lily_context.command_name == 'MAKE_IT'
        assert request._lily_context.correlation_id is not None

    def test_conf__is_saved_on_function(self):

        source = TestCommands.post.command_conf.pop('source')
        assert TestCommands.post.command_conf == {
            'name': 'MAKE_IT',
            'method': 'post',
            'meta': Meta(
                title='hi there',
                description='it is made for all',
                domain=Domain(id='test', name='test management')),
            'access': Access(
                access_list=['PREMIUM', 'SUPER_PREMIUM']),
            'input': Input(body_parser=TestCommands.BodyParser),
            'output': Output(serializer=TestCommands.ClientSerializer),
            'is_atomic': False,
            'fn': TestCommands.post.command_conf['fn'],
        }

        assert source.filepath == '/tests/test_base/test_command.py'
        assert source.start_line == 117
        assert source.end_line == 131

    #
    # INPUT
    #
    def test_input__body_parsing(self):

        u = User.objects.create_user(username='jacky')
        request = Mock(
            body=dump_to_bytes({'name': 'John', 'age': 81}),
            META=get_auth_headers(u.id))
        c = TestCommands()

        response = c.post(request, 19)

        assert request.input.body == {'name': 'John', 'age': 81}
        assert response.status_code == 200

    #
    # RESPONSE VALIDATION
    #
    def test_response_valid(self):

        u = User.objects.create_user(username='jacky')
        meta = get_auth_headers(u.id)
        meta['SERVER_NAME'] = 'testserver'
        request = Mock(
            body=dump_to_bytes({'amount': 81}),
            log_authorizer={
                'user_id': u.id,
            },
            META=meta)
        c = TestCommands()

        response = c.put(request)

        assert response.status_code == 200
        assert to_json(response) == {
            '@type': 'simple',
            '@event': 'BROKEN',
            'amount': 81,
        }

    def test_broken_response_validation(self):

        u = User.objects.create_user(username='jacky')
        meta = get_auth_headers(u.id)
        meta['SERVER_NAME'] = 'testserver'
        request = Mock(
            body=dump_to_bytes({'not_amount': 81}),
            log_authorizer={
                'user_id': u.id,
            },
            META=meta)
        c = TestCommands()

        response = c.put(request)

        assert response.status_code == 400
        assert to_json(response) == {
            '@authorizer': {
                'user_id': u.id,
            },
            '@type': 'error',
            '@event': 'RESPONSE_DID_NOT_VALIDATE',
            'errors': {'amount': ['This field is required.']},
        }

    #
    # ATOMICITY
    #
    def test_atomicity__correct_exceptions_are_visible_for_the_roolback(self):

        class AtomicContext(ContextDecorator):

            exception = None

            def __init__(self, *args, **kwargs):
                pass

            def __enter__(self):
                pass

            def __exit__(self, exc_type, exc, exc_tb):
                self.__class__.exception = exc

        self.mocker.patch.object(transaction, 'atomic', AtomicContext)

        u = User.objects.create_user(username='jacky')
        request = Mock(
            log_authorizer={
                'user_id': u.id,
            },
            META=get_auth_headers(u.id))
        c = TestCommands()
        self.mocker.patch.object(c, 'some_stuff').side_effect = [
            # -- database error
            DatabaseError,

            # -- generic exception
            Exception('hi there'),

            # -- lily exception
            event.BrokenRequest(
                'ERROR!',
                context=event.Context(user_id=u.id, origin='SERVICE')
            ),
        ]

        # -- db error
        response = c.delete(request)

        assert response.status_code == 500
        assert to_json(response) == {
            '@event': 'DATABASE_ERROR_OCCURRED',
            '@type': 'error',
            '@authorizer': {
                'user_id': u.id,
            },

        }
        assert isinstance(AtomicContext.exception, DatabaseError)

        # -- generic error
        response = c.delete(request)

        assert response.status_code == 500
        assert to_json(response) == {
            '@event': 'GENERIC_ERROR_OCCURRED',
            '@type': 'error',
            'errors': ['hi there'],
            '@authorizer': {
                'user_id': u.id,
            },
        }
        assert isinstance(AtomicContext.exception, Exception)

        # -- not a success exception
        response = c.delete(request)

        assert response.status_code == 400
        assert to_json(response) == {
            '@event': 'ERROR!',
            '@type': 'error',
            '@authorizer': {
                'user_id': u.id,
            },
        }
        assert isinstance(AtomicContext.exception, event.BrokenRequest)

    def test_atomicity__success_response_is_not_visible_to_rollback(self):

        class AtomicContext(ContextDecorator):

            exception = None

            def __init__(self, *args, **kwargs):
                pass

            def __enter__(self):
                pass

            def __exit__(self, exc_type, exc, exc_tb):
                self.__class__.exception = exc

        self.mocker.patch.object(transaction, 'atomic', AtomicContext)

        u = User.objects.create_user(username='jacky')
        request = Mock(
            user_id=u.id,
            email=None,
            origin=None,
            META=get_auth_headers(u.id))
        c = TestCommands()
        self.mocker.patch.object(c, 'some_stuff').side_effect = [

            # -- lily success exception
            event.Executed(
                event='SUCCESS!',
                context=event.Context(user_id=u.id)
            ),

            # -- lily created exception
            event.Created(
                event='CREATED!',
                context=event.Context(user_id=u.id)
            ),

        ]

        # -- success exception
        response = c.delete(request)

        assert response.status_code == 200
        assert to_json(response) == {'@event': 'SUCCESS!', '@type': 'empty'}
        assert AtomicContext.exception is None

        # -- created exception
        response = c.delete(request)

        assert response.status_code == 201
        assert to_json(response) == {'@event': 'CREATED!', '@type': 'empty'}
        assert AtomicContext.exception is None

    #
    # GENERIC ERRORS
    #
    def test_validation_error(self):

        u = User.objects.create_user(username='jacky')
        request = Mock(
            log_authorizer={
                'user_id': u.id,
            },
            META=get_auth_headers(u.id))
        c = TestCommands()
        self.mocker.patch.object(c, 'some_stuff').side_effect = (
            ValidationError({'field': ['is broken']}))

        response = c.delete(request)

        assert response.status_code == 400
        assert to_json(response) == {
            '@event': 'BODY_JSON_DID_NOT_PARSE',
            '@type': 'error',
            '@authorizer': {
                'user_id': u.id,
            },
            'errors': {'field': ['is broken']},
        }

    def test_does_not_exist(self):

        u = User.objects.create_user(username='jacky')
        request = Mock(
            log_authorizer={
                'user_id': u.id,
            },
            META=get_auth_headers(u.id))
        c = TestCommands()
        self.mocker.patch.object(c, 'some_stuff').side_effect = (
            lambda: User.objects.get(id=4930))

        response = c.delete(request)

        assert response.status_code == 404
        assert to_json(response) == {
            '@event': 'COULD_NOT_FIND_USER',
            '@type': 'error',
            '@authorizer': {
                'user_id': u.id,
            },
        }

    def test_multiple_objects_returned(self):

        u = User.objects.create_user(username='jacky')
        u = User.objects.create_user(username='jacks')
        request = Mock(
            log_authorizer={
                'user_id': u.id,
            },
            META=get_auth_headers(u.id))
        c = TestCommands()
        self.mocker.patch.object(c, 'some_stuff').side_effect = (
            lambda: User.objects.get(username__contains='jack'))

        response = c.delete(request)

        assert response.status_code == 500
        assert to_json(response) == {
            '@event': 'FOUND_MULTIPLE_INSTANCES_OF_USER',
            '@type': 'error',
            '@authorizer': {
                'user_id': u.id,
            },
        }

    #
    # GENERIC RESPONSE
    #
    def test_generic_response(self):

        # -- 201
        request = Mock(GET={'status_code': '201'}, META={})
        c = GenericCommands()

        response = c.get(request)

        assert response.status_code == 201
        assert to_json(response) == {
            '@event': 'FOUND_IT',
            '@type': 'what',
            'hi': 201,
        }

        # -- 404
        request = Mock(GET={'status_code': '404'}, META={})
        c = GenericCommands()

        response = c.get(request)

        assert response.status_code == 404
        assert to_json(response) == {
            '@event': 'FOUND_IT',
            '@type': 'what',
            'hi': 404,
        }

    #
    # HTTP RESPONSE
    #
    def test_http_response(self):

        # -- 404
        request = Mock(GET={'status_code': '404'}, META={})
        c = HttpCommands()

        response = c.get(request)

        assert response.status_code == 200
        assert response.content == b'hello world'
        assert response.get('Content-Type') == 'text/html; charset=utf-8'




from unittest.mock import Mock, call

from django.test import TestCase
import pytest

import search
from search import TextVector


def test_stored_vector__to_sql(mocker):

    compiler, connection = Mock(), Mock()
    as_sql = mocker.patch('search.vector.SearchVector.as_sql')

    vector = search.StoredVector('column')
    result = vector.as_sql(compiler, connection)

    assert (
        as_sql.call_args_list ==
        [call(
            compiler,
            connection,
            function=None,
            template='%(expressions)s')])
    assert result == as_sql.return_value


class TextVectorTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker):
        self.mocker = mocker
        self.mocker.patch('search.vector.HASHTAG_ESCAPE_SEQUENCE', '007')

    def setUp(self):
        self.vector = TextVector()

    def transform_exected(self, expected):

        if expected is None:
            return None

        return {
            k: set([str(e) for e in v])
            for k, v in expected.items()
        }

    def assert_parse(self, conf, text, weight, expected):

        stems = self.vector.parse(conf, text, weight)

        if stems is None:
            assert stems == expected

        else:
            assert stems == self.transform_exected(expected)

    #
    # PARSE - SIMPLE CONF
    #
    def test_parse__none_input_none_output(self):
        self.assert_parse('simple', None, None, {})

    def test_parse__empty_string_input_none_output(self):
        self.assert_parse('simple', '  ', None, {})

    def test_parse__conf_simple__single_word(self):
        self.assert_parse(
            'simple',
            'hello',
            None,
            {
                'hello': (1,),
            })

    def test_parse__conf_simple__many_words(self):
        self.assert_parse(
            'simple',
            'hi you how are you?',
            None,
            {
                'are': (4,),
                'hi': (1,),
                'how': (3,),
                'you': (2, 5),
            })

    def test_parse__conf_simple__with_single_hashtag(self):
        self.assert_parse(
            'simple',
            'hi #enjoyTheSilence how are you?',
            None,
            {
                '007enjoythesilence': (2,),
                'are': (4,),
                'hi': (1,),
                'how': (3,),
                'you': (5,),
            })

    def test_parse__conf_simple__with_many_hashtags(self):
        self.assert_parse(
            'simple',
            'hi #enjoyTheSilence how are #YouWin #watIsThis?',
            None,
            {
                'hi': (1,),
                '007enjoythesilence': (2,),
                'how': (3,),
                'are': (4,),
                '007youwin': (5,),
                '007watisthis': (6,),
            })

    def test_parse__conf_simple__transform_latex(self):
        self.assert_parse(
            'simple',
            r'#enjoyTheSilence $$ \frac{a}{b^2} $$',
            None,
            {
                '007enjoythesilence': (1,),
                'a': (3,),
                'by': (5,), 'b': (6,),
                'divided': (4,),
                'fraction': (2,),
                'squared': (7,),
            })

    #
    # PARSE - POLISH CONF
    #
    def test_parse__conf_polish__escape_accents(self):
        self.assert_parse(
            'polish',
            'część grzegżółek dość Świata',
            None,
            {
                'czesc': (1,),
                'część': (1,),
                'grzegzolek': (2,),
                'grzegżółek': (2,),
                'swiata': (4,),
                'świata': (4,),
            })

    def test_parse__conf_polish__escape_accents_triangulation(
            self):
        self.assert_parse(
            'polish',
            'Pchnąć w tę łódź jeża lub ośm skrzyń fig',
            None,
            {
                'figa': (9,),
                'figi': (9,),
                'jez': (5,),
                'jeza': (5,),
                'jeż': (5,),
                'lodz': (4,),
                'osm': (7,),
                'ośm': (7,),
                'pchnac': (1,),
                'pchnąć': (1,),
                'skrzyn': (8,),
                'skrzynia': (8,),
                'łódź': (4,),
            })

    def test_parse__conf_polish__weird_polish_characters_encoding(
            self):
        self.assert_parse(
            'polish',
            # -- warning: not visible to you eyes maybe but
            # -- I've manually added here crazy polish letters
            'Pchnąć w tę łódź jeża lub ośm skrzyń fig łódź.',
            None,
            {
                'figa': (9,),
                'figi': (9,),
                'jez': (5,),
                'jeza': (5,),
                'jeż': (5,),
                'lodz': (4, 10),
                'osm': (7,),
                'ośm': (7,),
                'pchnac': (1,),
                'pchnąć': (1,),
                'skrzyn': (8,),
                'skrzynia': (8,),
                'łódź': (4, 10),
            })

    def test_parse__conf_polish__many_hashtags(self):
        self.assert_parse(
            'polish',
            'część grzegżółek #dośćŚwiata #wokółNAS słyszałeś',
            None,
            {
                '007doscswiata': (3,),
                '007dośćświata': (3,),
                '007wokolnas': (4,),
                '007wokółnas': (4,),
                'czesc': (1,),
                'część': (1,),
                'grzegzolek': (2,),
                'grzegżółek': (2,),
                'słyszeć': (5,),
                'slyszec': (5,),
                'slyszales': (5,),
            })

    def test_parse__conf_polish__latex_transformation(self):
        self.assert_parse(
            'polish',
            'rozwiązanie to $$ \\sqrt{a ^ 2} $$',
            None,
            {
                'funkcja': (4,),
                'kwadrat': (10,),
                'kwadratowy': (6,),
                'pierwiastek': (5,),
                'pierwiastka': (5,),
                'rozwiazanie': (1,),
                'rozwiązanie': (1,),
                'rozwiazac': (1,),
                'rozwiązać': (1,),
                'sqrt': (3,),
            })

    def test_parse__conf_polish__with_weight(self):
        self.assert_parse(
            'polish',
            'dzień dobry świecie',
            'B',
            {
                'dobry': ('2B',),
                'dzien': ('1B',),
                'dzień': ('1B',),
                'swiecie': ('3B',),
                'świecie': ('3B',),
            })

    def test_parse__conf_polish__with_commas_and_dots(self):
        self.assert_parse(
            'polish',
            'rozwiązanie, to. jest, ok, ale, inne, jest prostsze.',
            None,
            {
                'prostszy': (8,),
                'rozwiazac': (1,),
                'rozwiazanie': (1,),
                'rozwiązanie': (1,),
                'rozwiązać': (1,),
            })

    def test_parse__conf_polish__with_other_extra_characters(self):
        self.assert_parse(
            'polish',
            'rozwiązanie;+- to%^& jest, ok, ale, san-jose, jest prostsze.',
            None,
            {
                'prostszy': (10,),
                'rozwiazac': (1,),
                'rozwiazanie': (1,),
                'rozwiązanie': (1,),
                'rozwiązać': (1,),
                'san-jose': (6,),
                'san': (7,),
                'jose': (8,),
            })

    def test_parse__conf_polish__multiple_positions(self):
        self.assert_parse(
            'polish',
            'adam adam ma kota kot ma adama',
            None,
            {
                'adam': set([2, 1]),
                'adama': set([7]),
                'kot': set([4, 5]),
                'kota': set([4, 5]),
            })

    def test_parse__conf_polish__broken_no_stem_words_positions(
            self):
        self.assert_parse(
            'polish',
            'zjezdzalnia brakuje polskich znakow',
            None,
            {
                'brakowac': (2,),
                'brakować': (2,),
                'brakuje': (2,),
                'polski': (3,),
                'zjezdzalnia': (1,),
                'znakow': (4,),
            })

    def test_parse__conf_polish__stop_words_removed(self):
        self.assert_parse(
            'polish',
            'i lub ale to ta te',
            None,
            {})

    def test_parse__conf_polish__not_have_to_contain_letters(self):
        self.assert_parse(
            'polish',
            '123 14 16 19 wynik to 190',
            None,
            {
                '123': (1,),
                '14': (2,),
                '16': (3,),
                '19': (4,),
                '190': (7,),
                'wynik': (5,)
            })

    #
    # PARSE - ENGLISH CONF
    #
    def test_parse__conf_english__many_hashtags(self):
        self.assert_parse(
            'english',
            'hello #around #theWorld',
            None,
            {
                'hello': (1,),
                '007around': (2,),
                '007theworld': (3,),
            })

    def test_parse__conf_english__latex_transformation(self):
        self.assert_parse(
            'english',
            'hello $$ \\frac{a}{b} $$',
            None,
            {
                'b': (6,),
                'divid': (4,),
                'fraction': (2,),
                'hello': (1,),
            })

    def test_parse__conf_english__with_weight(self):
        self.assert_parse(
            'english',
            'hello world',
            'A',
            {
                'hello': ('1A',),
                'world': ('2A',),
            })

    def test_parse__conf_english__with_commas_and_dots(self):
        self.assert_parse(
            'english',
            'hello, world. What, is. happening',
            None,
            {
                'happen': (5,),
                'hello': (1,),
                'world': (2,),
            })

    def test_parse__conf_english__with_other_extra_characters(
            self):
        self.assert_parse(
            'english',
            'hello;;*( world; What^& is. happening in san-jose',
            None,
            {
                'happen': (5,),
                'hello': (1,),
                'world': (2,),
                'jose': (9,),
                'san': (8,),
                'san-jos': (7,),
            })

    def test_parse__conf_english__multiple_positions(self):
        self.assert_parse(
            'english',
            'hello hello world hello what world?',
            None,
            {
                'hello': (4, 2, 1),
                'world': (6, 3),
            })

    def test_parse__conf_english__broken_no_stem_words_positions(
            self):
        self.assert_parse(
            'english',
            'helloooo worrld',
            None,
            {
                'helloooo': (1,),
                'worrld': (2,),
            })

    def test_parse__conf_english__stop_words_removed(self):
        self.assert_parse(
            'english',
            'who you are how is now?',
            None,
            {})

    def test_parse__conf_english__not_have_to_contain_letters(
            self):
        self.assert_parse(
            'english',
            'the day is 1893 and it 19 and 56',
            None,
            {
                '1893': (4,),
                '19': (7,),
                '56': (9,),
                'day': (2,),
            })

    #
    # TRANSFORM_SPECIAL_CHARACTERS
    #
    def test_transform_special_characters(self):

        assert self.vector.transform_special_characters(
            # -- warning: not visible to you eyes maybe but
            # -- I've manually added here crazy polish letters
            'Pchnąć w tę łódź jeża lub ośm skrzyń fig łódź.'
        ) == (
            'Pchnąć w tę łódź jeża lub ośm skrzyń fig łódź.'
        )

    def test_transform_special_characters__unicode(self):

        assert self.vector.transform_special_characters(
            # -- warning: not visible to you eyes maybe but
            # -- I've manually added here crazy polish letters
            'Pchna\u0328c\u0301 w te\u0328 \u0142o\u0301dz\u0301 '
            'jez\u0307a lub os\u0301m skrzyn\u0301 fig '
            '\u0142o\u0301dz\u0301.'
        ) == (
            'Pchnąć w tę łódź jeża lub ośm skrzyń fig łódź.'
        )

    #
    # PARSE_TO_LIST
    #
    def test_parse_to_list__makes_the_right_calls(self):

        parse = self.mocker.patch.object(TextVector, 'parse')
        parse.return_value = {
            'dzień': set(['1', '3']),
            'świecie': set(['2']),
        }
        assert self.vector.parse_to_list(
            'polish',
            'dzień dobry świecie',
            None
        ) == ['dzień', 'świecie']
        assert parse.call_args_list == [
            call('polish', 'dzień dobry świecie', None)
        ]

    #
    # PARSE_TO_TSVECTOR
    #
    def test_parse_to_tsvector__makes_the_right_calls(self):

        parse = self.mocker.patch.object(TextVector, 'parse')
        parse.return_value = {
            'dzień': set(['1']),
            'świecie': set(['2']),
        }
        assert self.vector.parse_to_tsvector(
            'polish',
            'dzień dobry świecie',
            None
        ) == "'dzień':1 'świecie':2"
        assert parse.call_args_list == [
            call('polish', 'dzień dobry świecie', None)
        ]

    #
    # TRANSFORM_HASHTAGS
    #
    def test_transform_hashtags(self):

        # -- single hashtag only
        assert self.vector.transform_hashtags('#hi') == '007hi'

        # -- single hashtag - at the beginning
        assert self.vector.transform_hashtags('#hi world') == '007hi world'

        # -- single hashtag - at the end
        assert self.vector.transform_hashtags('world #hi') == 'world 007hi'

        # -- single hashtag - in the middle
        assert self.vector.transform_hashtags(
            'world #hi you') == 'world 007hi you'

        # -- many hashtags
        assert self.vector.transform_hashtags(
            '#hi #what is not #this but #that'
        ) == '007hi 007what is not 007this but 007that'

        # -- unicode hashtags
        assert self.vector.transform_hashtags(
            '#ówdzie #późno is not #this but #that'
        ) == '007ówdzie 007późno is not 007this but 007that'

    #
    # TRANSFORM_LATEX
    #
    def test_transform_latex__conf_english(self):

        # -- single latex only
        assert self.vector.transform_latex(
            'english', '$$ \\frac{a}{b} $$'
        ) == 'fraction a divided by b'

        # -- single latex - at the beginning
        assert self.vector.transform_latex(
            'english', '$$ \\frac{a}{b} $$ and else'
        ) == 'fraction a divided by b and else'

        # -- single latex - at the end
        assert self.vector.transform_latex(
            'english', 'what is $$ \\frac{a}{b} $$'
        ) == 'what is fraction a divided by b'

        # -- single latex - in the middle
        assert self.vector.transform_latex(
            'english', 'what is $$ \\frac{a}{b} $$ and final'
        ) == 'what is fraction a divided by b and final'

        # -- many latex expressions
        assert self.vector.transform_latex(
            'english',
            'what is $$ \\frac{a}{b} $$ and $$ a $$ '
            'what $$ \\sqrt{x} $$'
        ) == (
            'what is fraction a divided by b and a what sqrt square root '
            'function of x'
        )

    def test_transform_latex__conf_polish(self):

        # -- single latex only
        assert self.vector.transform_latex(
            'polish', '$$ \\frac{a}{b} $$'
        ) == 'ułamek a dzielone przez b'

        # -- single latex - at the beginning
        assert self.vector.transform_latex(
            'polish', '$$ \\frac{a}{b} $$ i inne'
        ) == 'ułamek a dzielone przez b i inne'

        # -- single latex - at the end
        assert self.vector.transform_latex(
            'polish', 'czym jest $$ \\frac{a}{b} $$'
        ) == 'czym jest ułamek a dzielone przez b'

        # -- single latex - in the middle
        assert self.vector.transform_latex(
            'polish', 'czym jest $$ \\frac{a}{b} $$ i na koniec'
        ) == 'czym jest ułamek a dzielone przez b i na koniec'

        # -- many latex expressions
        assert self.vector.transform_latex(
            'polish',
            'czym jest $$ \\frac{a}{b} $$ oraz $$ a $$ '
            'czym $$ \\sqrt{x} $$'
        ) == (
            'czym jest ułamek a dzielone przez b oraz a czym sqrt funkcja '
            'pierwiastek kwadratowy z x'
        )

    #
    # TOKENIZE
    #
    def test_tokenize(self):

        # -- empty case
        assert self.vector.tokenize('simple', '') == []

        # -- conf simple
        assert self.vector.tokenize('simple', 'hi there how are you') == [
            ('hi', 1), ('there', 2), ('how', 3), ('are', 4), ('you', 5)
        ]

        # -- conf english
        assert self.vector.tokenize('english', 'hi there how are you') == [
            ('hi', 1), ('there', 2), ('how', 3), ('are', 4), ('you', 5)
        ]

        # -- conf polish
        assert self.vector.tokenize(
            'polish',
            'cześć wam jak się czujecie?'
        ) == [
            ('cześć', 1), ('czujecie', 5)
        ]

        # -- extra characters are removed
        assert self.vector.tokenize('simple', 'hi;; there+=& how are you') == [
            ('hi', 1), ('there', 2), ('how', 3), ('are', 4), ('you', 5)
        ]

    #
    # AUGUMENT_WITH_UNACCENTS
    #
    def test_augument_with_unaccents__no_accents(self):

        assert self.vector.augument_with_unaccents(
            'polish',
            None,
            {'mam': set(['3'])},
            [('hej', 1), ('jak', 2), ('macie', 3)]
        ) == {
            'mam': set(['3']),
        }

    def test_augument_with_unaccents__tokens_only(self):

        assert self.vector.augument_with_unaccents(
            'polish',
            None,
            {},
            [('cześć', 1), ('jak', 2), ('świecie', 3)]
        ) == {
            'czesc': set(['1']),
            'swiecie': set(['3']),
        }

    def test_augument_with_unaccents__stems_only(self):

        assert self.vector.augument_with_unaccents(
            'polish',
            None,
            {
                'cześć': set(['1']),
                'jak': set(['2']),
                'świecie': set(['3', '4']),
            },
            []
        ) == {
            'czesc': set(['1']),
            'cześć': set(['1']),
            'jak': set(['2']),
            'świecie': set(['3', '4']),
            'swiecie': set(['3', '4']),
        }

    def test_augument_with_unaccents__stems_and_tokens(self):

        assert self.vector.augument_with_unaccents(
            'polish',
            None,
            {'cześć': set(['1']), 'czesc': set(['2'])},
            [('cześć', 1), ('jak', 2), ('macie', 3)]
        ) == {
            'czesc': set(['1', '2']),
            'cześć': set(['1']),
        }

    #
    # AUGUMENT_WITH_STEMS
    #
    def test_augument_with_stems__simple_conf(self):

        assert self.vector.augument_with_stems(
            'simple',
            'hello there',
            None,
            {
                'already': set(['1', '5']),
            },
            [('will', 1), ('be', 2), ('ignored', 3)]
        ) == {
            'already': set(['1', '5']),
            'hello': set(['1']),
            'there': set(['2']),
        }

    def test_augument_with_stems__polish_conf(self):

        assert self.vector.augument_with_stems(
            'polish',
            'cześć wam zdobyliście wszystkie zasługi?',
            None,
            {},
            [
                ('cześć', 1),
                ('zdobyliście', 3),
                ('wszystkie', 4),
                ('zasługi', 5),
            ]
        ) == {
            'cześć': set(['1']),
            'wszystkie': set(['4']),
            'zasługa': set(['5']),
            'zdobyć': set(['3']),
        }

    def test_augument_with_stems__english_conf(self):

        assert self.vector.augument_with_stems(
            'english',
            'hey there did you gained everything?',
            None,
            {},
            [('will', 1), ('be', 2), ('ignored', 3)]
        ) == {
            'everyth': set(['6']),
            'gain': set(['5']),
            'hey': set(['1']),
        }


#
# HELPERS
#
@pytest.mark.parametrize("v1, v2, v3, expected", [

    # case 0 - single word vectors
    ("'cat':1", "'dog':1", None, "'cat':1 'dog':2"),

    # case 1 - single word vectors with weights
    ("'cat':1B", "'dog':1C", None, "'cat':1B 'dog':2C"),

    # case 2 - many words
    (
        "'cat':1 'whatever':11",
        "'dog':9 'hospital':11 'life':13",
        None,
        "'cat':1 'whatever':11 'dog':20 'hospital':22 'life':24",
    ),

    # case 3 - many words with weights
    (
        "'cat':1B 'whatever':11C",
        "'dog':9 'hospital':11A 'life':13",
        None,
        "'cat':1B 'whatever':11C 'dog':20 'hospital':22A 'life':24",
    ),

    # case 4 - three vectors
    (
        "'cat':1B 'whatever':11C",
        "'dog':9 'hospital':11A 'life':13",
        "'whatever':1B",
        "'cat':1B 'whatever':11C,25B 'dog':20 'hospital':22A 'life':24",
    ),

], ids=list(map(str, range(0, 5))))
@pytest.mark.django_db
def test_concatenate(v1, v2, v3, expected):

    def transform(x):
        if not x:
            return x

        return set(x.split(' '))

    if v3:
        vectors = [v1, v2, v3]

    else:
        vectors = [v1, v2]

    assert (
        transform(search.concatenate_tsvectors(*vectors)) ==
        transform(expected))




import re

from .utils import to_camelcase


class Signature:

    def __init__(
            self,
            method,
            path,
            path_parameters,
            authorization_required=True,
            request_query=None,
            request_body=None,
            bulk_read_field=None):
        self.method = method
        self.path = path
        self.path_parameters = path_parameters
        self.authorization_required = authorization_required
        self.request_query = request_query
        self.request_body = request_body
        self.bulk_read_field = bulk_read_field

    @property
    def input(self):

        input_signature = []
        # -- path parameters
        if self.path_parameters:
            input_signature.append(
                ', '.join([self.to_arg(p) for p in self.path_parameters]))

        # -- query
        if not self.request_query.is_empty():
            input_signature.append(
                'params: X.{}'.format(self.request_query.name))

        # -- body
        if not self.request_body.is_empty():
            input_signature.append(
                'body: X.{}'.format(self.request_body.name))

        return ', '.join(input_signature)

    @property
    def call_args(self):

        call_args_signature = []

        # -- path parameters
        path = re.sub(
            r'\{(?P<parameter>[\w\_]+)\}',
            lambda x: '${{{}}}'.format(
                to_camelcase(x.groupdict()['parameter'], first_lower=True)),
            self.path)

        if self.path_parameters:
            call_args_signature.append('`{}`'.format(path))

        else:
            call_args_signature.append("'{}'".format(path))

        #
        # BODY
        #
        # -- body
        if not self.request_body.is_empty():
            call_args_signature.append('body')

        # -- when body is missing
        elif self.method in ['put', 'post']:
            call_args_signature.append('{}')

        #
        # OPTIONS (INCLUDING QUERY PARAMS)
        #
        # -- query
        options = []
        if not self.request_query.is_empty():
            options.append('params')

        # -- authorization
        options.append(
            "authorizationRequired: {}".format(
                self.authorization_required and 'true' or 'false'))

        call_args_signature.append('{{ {} }}'.format(', '.join(options)))

        return ', '.join(call_args_signature)

    @property
    def call_args_without_path(self):

        call_args_signature = []

        # -- path parameters
        if self.path_parameters:
            call_args_signature.append(
                ', '.join([
                    to_camelcase(p['name'], first_lower=True)
                    for p in self.path_parameters]))

        # -- query
        if not self.request_query.is_empty():
            call_args_signature.append('params')

        # -- body
        if not self.request_body.is_empty():
            call_args_signature.append('body')

        return ', '.join(call_args_signature)

    def to_arg(self, parameter):
        if parameter['type'] == int:
            type_ = 'number'

        elif parameter['type'] == str:
            type_ = 'string'

        else:
            type_ = 'any'

        return '{arg}: {type}'.format(
            arg=to_camelcase(parameter['name'], first_lower=True),
            type=type_)




from unittest.mock import call

from django.test import TestCase, override_settings
from django.urls import re_path, reverse
import pytest
from conf.urls import urlpatterns

from lily.base.test import Client
from lily.base.commands import S3UploadSignCommands
from lily.base.command import command_override
from lily import (
    Meta,
    name,
    Domain,
    Access,
)


MySignCommands = S3UploadSignCommands.overwrite(
    get=command_override(
        name=name.Execute('SIGN', 'PROCESS'),
        meta=Meta(
            title=(
                'Sign Process dedicated to upload and conversion of media '
                'file'),
            domain=Domain(id='hey', name='hi')),
        access=Access(
            access_list=['PREMIUM', 'SUPER_PREMIUM'])))


urlpatterns.extend([
    re_path(r'^sign/$', MySignCommands.as_view(), name='test.sign'),

])


class MySignCommandsTestCase(TestCase):

    uri = reverse('test.sign')

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker):
        self.mocker = mocker

    def setUp(self):
        self.app = Client()
        self.user_id = 56
        self.auth_headers = {
            'HTTP_X_CS_ACCOUNT_TYPE': 'PREMIUM',
            'HTTP_X_CS_USER_ID': self.user_id,
        }

    @override_settings(AWS_S3_ACCESS_KEY="access.key", AWS_S3_REGION="central")
    def test_get_200(self):

        sign_mock = self.mocker.patch.object(MySignCommands, 'sign')
        sign_mock.side_effect = [
            'signature 1', 'signature 2', 'signature 3', 'signature 4',
        ]
        get_signature_mock = self.mocker.patch.object(
            MySignCommands, 'get_signature')
        get_signature_mock.return_value = "af52522c5afb83b5348ed06b5fbd0c"

        response = self.app.get(
            self.uri,
            data={
                "to_sign": "hi there",
                "datetime": "20171201T123112Z",
            },
            **self.auth_headers)

        assert response.status_code == 200
        assert response.content == b"af52522c5afb83b5348ed06b5fbd0c"
        assert (
            get_signature_mock.call_args_list ==
            [call('signature 4', 'hi there')])
        assert (
            sign_mock.call_args_list == [
                call(b'AWS4access.key', '20171201'),
                call('signature 1', 'central'),
                call('signature 2', 's3'),
                call('signature 3', 'aws4_request'),
            ])

    def test_get_400__missing_fields(self):

        # -- missing to sign
        response = self.app.get(
            self.uri,
            data={"datetime": "20171201T123112Z"},
            **self.auth_headers)

        assert response.status_code == 400
        assert response.json() == {
            '@event': 'QUERY_DID_NOT_VALIDATE',
            '@type': 'error',
            'errors': {'to_sign': ['This field is required.']},
            '@authorizer': {
                'account_type': 'PREMIUM',
                'user_id': self.user_id,
            }
        }

        # -- missing datetime
        response = self.app.get(
            self.uri,
            data={"to_sign": "hi there"},
            **self.auth_headers)

        assert response.status_code == 400
        assert response.json() == {
            '@event': 'QUERY_DID_NOT_VALIDATE',
            '@type': 'error',
            'errors': {'datetime': ['This field is required.']},
            '@authorizer': {
                'account_type': 'PREMIUM',
                'user_id': self.user_id,
            },
        }

    def test_get_400__broken_datetime(self):

        response = self.app.get(
            self.uri,
            data={
                "to_sign": "hi there",
                "datetime": "2017000001201T123112Z",
            },
            **self.auth_headers)

        assert response.status_code == 400
        assert response.json() == {
            '@event': 'QUERY_DID_NOT_VALIDATE',
            '@type': 'error',
            'errors': {
                'datetime': [(
                    'invalid datetime format accepted is YYYYMMDDThhmmssZ'
                )]
            },
            '@authorizer': {
                'account_type': 'PREMIUM',
                'user_id': self.user_id,
            },
        }




# import re

from django.contrib.postgres.search import SearchQuery

from .detector import detector
# from .latex.transformer import transform
from .vector import TextVector
from .constants import HASHTAG_ESCAPE_SEQUENCE, HASHTAG_PATTERN


class Query(SearchQuery):
    """Query object.

    Query object with auto-language detection feature and latex syntax
    parser.

    """

    forbidden_chars = '!"$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'

    def __init__(self, value, output_field=None, **extra):
        language_conf = detector.detect_db_conf(value)
        self.parsed_value = self.parse_value(value, language_conf)

        super(Query, self).__init__(value, config=language_conf)

    def parse_value(self, text, language_conf):

        # table = str.maketrans(
        #     self.forbidden_chars, len(self.forbidden_chars) * " ")

        # text = transform(text, language_conf)
        # text = text.translate(table).strip()

        # # fetch all hashtags
        hashtags = []

        def remove_hashtag(match):
            hashtags.append(
                HASHTAG_ESCAPE_SEQUENCE + match.group('text'))

            return match.group('prefix') + ' '

        tokens_only_text = HASHTAG_PATTERN.sub(remove_hashtag, text)
        tokens = TextVector().parse_to_list(language_conf, tokens_only_text)

        # if not tokens_only_text.strip():
        #     tokens = []

        # else:
        #     tokens = re.split(r'\s+', tokens_only_text.strip())

        if hashtags and tokens:
            return '{hashtags} & ({tokens})'.format(
                hashtags=' & '.join(hashtags),
                tokens=' | '.join(tokens))

        elif hashtags and not tokens:
            return ' & '.join(hashtags)

        elif not hashtags and tokens:
            return ' | '.join(tokens)

    def as_sql(self, compiler, connection):
        params = [self.parsed_value]

        config_sql, config_params = compiler.compile(self.config)
        template = 'to_tsquery({}::regconfig, %s)'.format(config_sql)
        params = config_params + params

        return template, params




from unittest import TestCase
from unittest.mock import call
import json
import os
import textwrap

import pytest

from lily.cli.copier import Copier


class CopierTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def init_fixtures(self, tmpdir, mocker):
        self.tmpdir = tmpdir
        self.mocker = mocker

    def setUp(self):
        self.cli_dir = self.tmpdir.mkdir('cli_dir')
        self.project_dir = self.tmpdir.mkdir('project')
        self.project_dir.mkdir('.lily').join('config.json').write(json.dumps({
            'version': '0.0.11'
        }))
        self.current_cwd = os.getcwd()
        os.chdir(str(self.project_dir))

    def tearDown(self):
        os.chdir(self.current_cwd)

    #
    # COPY
    #
    def test_copy__makes_the_right_calls(self):

        copy_makefile = self.mocker.patch.object(Copier, 'copy_makefile')

        Copier().copy('gigly')

        assert copy_makefile.call_args_list == [call('gigly')]

    #
    # COPY_MAKEFILE
    #
    def test_copy_makefile__copies_makefile(self):

        makefile = self.cli_dir.join('base.makefile')
        makefile.write(textwrap.dedent('''
            ## GENERATED FOR VERSION: {% VERSION %}

            lint:  ## lint the {% SRC_DIR %} & tests
                source env.sh && \
                flake8 --ignore D100,D101 tests && \
                flake8 --ignore D100,D101 {% SRC_DIR %}
        '''))
        self.mocker.patch.object(Copier, 'base_makefile_path', str(makefile))

        Copier().copy_makefile('gigly')

        result_makefile_content = (
            self.project_dir.join('.lily/lily.makefile').read())
        assert result_makefile_content == textwrap.dedent('''
            ## GENERATED FOR VERSION: 0.0.11

            lint:  ## lint the gigly & tests
                source env.sh && \
                flake8 --ignore D100,D101 tests && \
                flake8 --ignore D100,D101 gigly
        ''')

    def test_copy_makefile__copies_makefile__with_override(self):

        makefile = self.cli_dir.join('base.makefile')
        makefile.write('NEW make it')
        self.mocker.patch.object(Copier, 'base_makefile_path', str(makefile))
        self.project_dir.join(
            '.lily/lily.makefile').write('OLD make it')

        Copier().copy_makefile('gigly')

        assert (
            self.project_dir.join('.lily/lily.makefile').read() ==
            'NEW make it')




import re
import os
import json
import logging
from collections import OrderedDict

from django.template import engines

from lily.entrypoint.base import BaseRenderer
from lily.base.meta import MetaSerializer
from lily.base.test import get_examples_filepath


BASE_DIR = os.path.dirname(__file__)


BASE_TEMPLATE_PATH = os.path.join(BASE_DIR, './base.md')


engine = engines['django']


logger = logging.getLogger('RENDERER')


def clean_extra_indents(x):
    return re.sub(r'\n +', r'\n', x)


def clean_extra_new_lines(x):
    return re.sub(r'\n+', r'\n', x)


def dump_or_none(x):
    if x:
        return json.dumps(x, indent=4, sort_keys=True)


class MarkdownRenderer(BaseRenderer):

    def __init__(self, urlpatterns):
        self.urlpatterns = urlpatterns

    def render(self):

        base_index = super(MarkdownRenderer, self).render()

        with open(BASE_TEMPLATE_PATH, 'r') as f:
            template = engine.from_string(f.read())

        examples = self.get_examples()
        document_tree = {}
        sorted_index = sorted(
            base_index.items(), key=lambda x: x[0])

        for command_name, conf in sorted_index:
            meta = MetaSerializer(conf['meta']).data

            domain = meta['domain']['name']
            document_tree.setdefault(domain, [])

            try:
                command_examples = examples[command_name]

            except KeyError:
                logger.error(
                    'Missing examples for {}'.format(command_name))

            else:
                document_tree[domain].append({
                    'title': '{name}: {method} {path}'.format(
                        name=command_name,
                        method=conf['method'].upper(),
                        path=conf['path_conf']['path']),
                    'description': clean_extra_indents(
                        '{title} \n {description}'.format(
                            title=meta['title'],
                            description=meta['description'])),
                    'calls': sorted([
                        {
                            'title': t,
                            'response': {
                                'content': dump_or_none(
                                    e['response']['content']),
                            },
                            'request': {
                                'method': conf['method'],
                                'path': e['request']['path'],
                                'content': dump_or_none(
                                    e['request'].get('content')),
                                'headers': self.render_headers(e)
                            }
                        }
                        for t, e in command_examples.items()
                    ], key=lambda x: x['title'])
                })

        return clean_extra_new_lines(template.render({
            'document_tree': OrderedDict(
                sorted(document_tree.items(), key=lambda x: x[0])),
        }))

    def render_headers(self, e):
        return OrderedDict(sorted(
            e['request'].get('headers', {}).items(),
            key=lambda x: x[0]))

    def get_examples(self):
        with open(get_examples_filepath()) as f:
            return json.loads(f.read())




import json
import os

from django.test import TestCase
from django.http import HttpResponse
from django.views.generic import View
from django.urls import re_path
import pytest

from conf.urls import urlpatterns

from lily.base.command import command
from lily.base.meta import Meta, Domain
from lily.base.access import Access
from lily.base.output import Output
from lily.base import serializers
from lily.base.test import Client


class HttpView(View):

    @command(
        name='HTTP_IT',
        meta=Meta(
            title='http it',
            description='http it',
            domain=Domain(id='d', name='d')),
        access=Access(access_list=[]))
    def get(self, request):
        return HttpResponse('http it!')


class SampleSerializer(serializers.Serializer):

    _type = 'test'

    hello = serializers.CharField()


class SampleView(View):

    @command(
        name='POST_IT',
        meta=Meta(
            title='post it',
            description='post it',
            domain=Domain(id='d', name='d')),
        access=Access(access_list=[]),
        output=Output(serializer=SampleSerializer))
    def post(self, request):
        raise self.event.Created(
            event='CREATED', context=request, data={'hello': 'post'})

    @command(
        name='GET_IT',
        meta=Meta(
            title='get it',
            description='get it',
            domain=Domain(id='d', name='d')),
        access=Access(access_list=[]),
        output=Output(serializer=SampleSerializer))
    def get(self, request):

        _type = request.GET.get('type')
        if _type in ['a', None]:
            raise self.event.Executed(
                event='LISTED', context=request, data={'hello': 'get.a'})

        if _type == 'b':
            raise self.event.Executed(
                event='LISTED', context=request, data={'hello': 'get.b'})

        if _type == 'c':
            raise self.event.Executed(
                event='LISTED_AGAIN', context=request, data={'hello': 'get.c'})

        if _type == 'd':
            raise self.event.DoesNotExist(
                event='ERROR_LISTED', context=request, data={'hello': 'get.d'})

    @command(
        name='PUT_IT',
        meta=Meta(
            title='put it',
            description='put it',
            domain=Domain(id='d', name='d')),
        access=Access(access_list=[]),
        output=Output(serializer=SampleSerializer))
    def put(self, request):
        raise self.event.Executed(
            event='UPDATED', context=request, data={'hello': 'put'})

    @command(
        name='DELETE_IT',
        meta=Meta(
            title='delete it',
            description='delete it',
            domain=Domain(id='d', name='d')),
        access=Access(access_list=[]),
        output=Output(serializer=SampleSerializer))
    def delete(self, request):
        raise self.event.Executed(
            event='DELETED', context=request, data={'hello': 'delete'})


urlpatterns.extend([
    re_path(r'^test/it/$', SampleView.as_view(), name='test.it'),
    re_path(r'^http/it/$', HttpView.as_view(), name='http.it'),
])


class ClientTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, mocker, tmpdir):
        self.tmpdir = tmpdir
        self.mocker = mocker

    def prepare_example_file(self):
        examples_file = self.tmpdir.mkdir('docs').join('examples.json')
        examples_file.write('{}')
        get_examples_filepath = self.mocker.patch(
            'lily.base.test.get_examples_filepath')
        get_examples_filepath.return_value = str(examples_file)

        return examples_file

    def test_create_file_if_does_not_exist(self):
        examples_file_dir = self.tmpdir.mkdir('docs')
        filepath = os.path.join(str(examples_file_dir), 'examples007.json')

        get_examples_filepath = self.mocker.patch(
            'lily.base.test.get_examples_filepath')
        get_examples_filepath.return_value = filepath

        response = Client().get('/test/it/')

        examples_file = open(filepath, 'r')
        assert response.status_code == 200
        assert json.loads(examples_file.read()) == {
            'GET_IT': {
                '200 (LISTED)': {
                    'method': 'get',
                    'description': 'LISTED',
                    'request': {
                        'path': '/test/it/',
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'LISTED',
                            'hello': 'get.a',
                        },
                    },

                },
            },
        }

    def test_single_response(self):
        examples_file = self.prepare_example_file()

        response = Client().get('/test/it/')

        assert response.status_code == 200
        assert json.loads(examples_file.read()) == {
            'GET_IT': {
                '200 (LISTED)': {
                    'method': 'get',
                    'description': 'LISTED',
                    'request': {
                        'path': '/test/it/',
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'LISTED',
                            'hello': 'get.a',
                        },
                    },

                },
            },
        }

    def test_http_response(self):
        examples_file = self.prepare_example_file()

        response = Client().get('/http/it/')

        assert response.status_code == 200
        assert json.loads(examples_file.read()) == {
            'HTTP_IT': {
                '200': {
                    'method': 'get',
                    'description': None,
                    'request': {
                        'path': '/http/it/',
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'text/html; charset=utf-8',
                        'content': 'http it!',
                    },

                },
            },
        }

    def test_with_extra_desc(self):
        examples_file = self.prepare_example_file()

        response = Client().get('/test/it/', extra_desc='special')

        assert response.status_code == 200
        assert json.loads(examples_file.read()) == {
            'GET_IT': {
                '200 (LISTED) - special': {
                    'method': 'get',
                    'description': 'LISTED',
                    'request': {
                        'path': '/test/it/',
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'LISTED',
                            'hello': 'get.a',
                        },
                    },

                },
            },
        }

    def test_multiple_different_responses_single_endpoint(self):
        examples_file = self.prepare_example_file()

        Client().get('/test/it/?type=a')
        Client().get('/test/it/?type=b')
        Client().get('/test/it/?type=c')
        Client().get('/test/it/?type=d')

        assert json.loads(examples_file.read()) == {
            'GET_IT': {
                '200 (LISTED)': {
                    'method': 'get',
                    'description': 'LISTED',
                    'request': {
                        'path': '/test/it/?type=b',
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'LISTED',
                            'hello': 'get.b',
                        },
                    }
                },
                '200 (LISTED_AGAIN)': {
                    'method': 'get',
                    'description': 'LISTED_AGAIN',
                    'request': {
                        'path': '/test/it/?type=c',
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'LISTED_AGAIN',
                            'hello': 'get.c',
                        },
                    },
                },
                '404 (ERROR_LISTED)': {
                    'method': 'get',
                    'description': 'ERROR_LISTED',
                    'request': {
                        'path': '/test/it/?type=d',
                    },
                    'response': {
                        'status': 404,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'error',
                            '@event': 'ERROR_LISTED',
                            'hello': 'get.d',
                        },
                    },
                },
            },
        }

    def test_multiple_same_responses_single_endpoint(self):
        examples_file = self.prepare_example_file()

        Client().get('/test/it/?type=a')
        Client().get('/test/it/?type=b')
        Client().get('/test/it/?type=c')
        Client().get('/test/it/?type=c')
        Client().get('/test/it/?type=b')
        Client().get('/test/it/?type=c')

        assert json.loads(examples_file.read()) == {
            'GET_IT': {
                '200 (LISTED)': {
                    'method': 'get',
                    'description': 'LISTED',
                    'request': {
                        'path': '/test/it/?type=b',
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'LISTED',
                            'hello': 'get.b',
                        },
                    }
                },
                '200 (LISTED_AGAIN)': {
                    'method': 'get',
                    'description': 'LISTED_AGAIN',
                    'request': {
                        'path': '/test/it/?type=c',
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'LISTED_AGAIN',
                            'hello': 'get.c',
                        },
                    },
                },
            },
        }

    def test_multiple_responses_for_multiple_endpoints(self):
        examples_file = self.prepare_example_file()

        Client().get('/test/it/?type=a')
        Client().get('/test/it/?type=b')
        Client().get('/test/it/?type=a')
        Client().delete('/test/it/')
        Client().get('/test/it/?type=d')
        Client().put('/test/it/')

        assert json.loads(examples_file.read()) == {
            'GET_IT': {
                '200 (LISTED)': {
                    'method': 'get',
                    'description': 'LISTED',
                    'request': {
                        'path': '/test/it/?type=a'
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'LISTED',
                            'hello': 'get.a',
                        },
                    }
                },
                '404 (ERROR_LISTED)': {
                    'method': 'get',
                    'description': 'ERROR_LISTED',
                    'request': {
                        'path': '/test/it/?type=d'
                    },
                    'response': {
                        'status': 404,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'error',
                            '@event': 'ERROR_LISTED',
                            'hello': 'get.d',
                        },
                    }
                },
            },
            'DELETE_IT': {
                '200 (DELETED)': {
                    'method': 'delete',
                    'description': 'DELETED',
                    'request': {
                        'path': '/test/it/'
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'DELETED',
                            'hello': 'delete',
                        },
                    },
                }
            },
            'PUT_IT': {
                '200 (UPDATED)': {
                    'method': 'put',
                    'description': 'UPDATED',
                    'request': {
                        'path': '/test/it/'
                    },
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'UPDATED',
                            'hello': 'put',
                        },
                    },
                }
            },
        }

    def test_body_examples_for_post_and_put(self):
        examples_file = self.prepare_example_file()

        Client().post(
            '/test/it/',
            data=json.dumps({'create': 'it'}),
            content_type='application/json')
        Client().put(
            '/test/it/',
            data=json.dumps({'update': 'it', 'please': 'now'}),
            content_type='application/json')

        assert json.loads(examples_file.read()) == {
            'POST_IT': {
                '201 (CREATED)': {
                    'method': 'post',
                    'description': 'CREATED',
                    'response': {
                        'status': 201,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'CREATED',
                            'hello': 'post',
                        }
                    },
                    'request': {
                        'path': '/test/it/',
                        'headers': {
                            'CONTENT-TYPE': 'application/json',
                        },
                        'content': {
                            'create': 'it',
                        }
                    }
                },
            },
            'PUT_IT': {
                '200 (UPDATED)': {
                    'method': 'put',
                    'description': 'UPDATED',
                    'response': {
                        'status': 200,
                        'content_type': 'application/json',
                        'content': {
                            '@type': 'test',
                            '@event': 'UPDATED',
                            'hello': 'put',
                        },
                    },
                    'request': {
                        'path': '/test/it/',
                        'headers': {
                            'CONTENT-TYPE': 'application/json',
                        },
                        'content': {
                            'update': 'it',
                            'please': 'now',
                        }
                    }
                }
            },
        }




from unittest.mock import Mock

from django.test import TestCase
from django.contrib.auth.models import User
import pytest

from lily.base import parsers
from lily.base.events import EventFactory
from lily.base.input import Input
from .test_parsers import RequestGet


class InputTestCase(TestCase):

    def setUp(self):
        User.objects.all().delete()

    @pytest.fixture(autouse=True)
    def initfitures(self, mocker):
        self.mocker = mocker

    #
    # PARSE
    #
    def test_parse__makes_the_right_calls(self):
        parse_body = self.mocker.patch.object(Input, 'parse_body')
        parse_body.return_value = [{}, {}]
        parse_query = self.mocker.patch.object(Input, 'parse_query')
        query_parser, body_parser = Mock(), Mock()
        request = Mock()

        # -- both query and body parsers
        i = Input(
            query_parser=query_parser, body_parser=body_parser)
        i.parse(request, command_name='MAKE_IT')

        assert parse_body.call_count == 1
        assert parse_query.call_count == 1

    #
    # PARSE_QUERY
    #
    def test_parse_query__all_ok(self):
        class QueryParser(parsers.Parser):
            title = parsers.CharField()
            prices = parsers.ListField(child=parsers.IntegerField())

        input = Input(query_parser=QueryParser)

        request = Mock(
            GET=RequestGet(title=['hi there'], prices=[67, 89, 11]),
            user_id=902)

        data = input.parse_query(request)

        assert data == {'title': 'hi there', 'prices': [67, 89, 11]}

    def test_parse_query__multiple_parsers(self):
        class AParser(parsers.Parser):
            title = parsers.CharField()
            prices = parsers.ListField(child=parsers.IntegerField())

        class BParser(parsers.Parser):
            title = parsers.CharField()
            quantity = parsers.IntegerField()

        class CParser(AParser, BParser):
            pass

        input = Input(query_parser=CParser)

        request = Mock(
            GET=RequestGet(
                title=['hi there'],
                prices=[67, 89, 11],
                quantity=[190]),
            user_id=902)

        data = input.parse_query(request)

        assert data == {
            'title': 'hi there',
            'prices': [67, 89, 11],
            'quantity': 190,
        }

    def test_parse_query__all_missing(self):
        class QueryParser(parsers.Parser):
            title = parsers.CharField()
            prices = parsers.ListField(child=parsers.IntegerField())

        input = Input(query_parser=QueryParser)

        request = Mock(
            GET=RequestGet(),
            origin=None,
            log_authorizer={
                'user_id': 902,
            })

        with pytest.raises(EventFactory.BrokenRequest) as e:
            input.parse_query(request)

        assert e.value.event == 'QUERY_DID_NOT_VALIDATE'

    def test_parse_query__event__query_did_not_validate(self):

        class QueryParser(parsers.Parser):
            title = parsers.CharField()

            prices = parsers.ListField(child=parsers.IntegerField())

        input = Input(query_parser=QueryParser)

        request = Mock(
            GET=RequestGet(title=['hi there'], prices=['what']),
            origin=None,
            log_authorizer={
                'user_id': 902,
            })

        with pytest.raises(EventFactory.BrokenRequest) as e:
            input.parse_query(request)

        assert e.value.data == {
            '@type': 'error',
            '@event': 'QUERY_DID_NOT_VALIDATE',
            '@authorizer': {
                'user_id': 902,
            },
            'errors': {
                'prices': {0: ['A valid integer is required.']},
            },
        }

    #
    # PARSE_BODY
    #
    def _prepare_body_parser(self):
        class BodyParser(parsers.Parser):
            title = parsers.CharField()
            amount = parsers.IntegerField(max_value=19)

        input = Input(body_parser=BodyParser)

        return input

    def test_parse_body__all_ok(self):
        input = self._prepare_body_parser()

        request = Mock(
            body=b'{"title": "hi there", "amount": "18"}', user_id=902)

        data, raw_data = input.parse_body(request, command_name='MAKE_IT')

        assert data == {'amount': 18, 'title': 'hi there'}
        assert raw_data == {'amount': '18', 'title': 'hi there'}

    def test_parse_body__fixing_broken_characters(self):
        input = self._prepare_body_parser()

        request = Mock(
            body=(
                # -- warning: not visible to you eyes maybe but
                # -- I've manually added here crazy polish letters
                '{"title": "Pchnąć w tę ło\u0301dź jeża lub ośm ' +
                'skrzyń fig", "amount": "12"}').encode('utf8'),
            origin=None,
            log_authorizer={
                'user_id': 902,
            })

        data, raw_data = input.parse_body(request, command_name='MAKE_IT')

        assert data == {
            'amount': 12,
            'title': 'Pchnąć w tę łódź jeża lub ośm skrzyń fig',
        }
        assert raw_data == {
            'amount': '12',
            'title': 'Pchnąć w tę łódź jeża lub ośm skrzyń fig',
        }

    def test_parse_body__broken_json(self):
        input = self._prepare_body_parser()

        request = Mock(
            body=b'{not json',
            origin=None,
            log_authorizer={
                'user_id': 902,
            })

        with pytest.raises(EventFactory.BrokenRequest) as e:
            input.parse_body(request, command_name='MAKE_IT')

        assert e.value.event == 'BODY_JSON_DID_NOT_PARSE'
        assert e.value.data == {
            '@type': 'error',
            '@event': 'BODY_JSON_DID_NOT_PARSE',
            '@authorizer': {
                'user_id': 902,
            },
        }
        assert e.value.is_critical

    def test_parse_body__invalid_payload(self):
        input = self._prepare_body_parser()

        # -- amount too big error
        request = Mock(
            body=b'{"title": "hi there", "amount": 20}',
            log_authorizer={
                'user_id': 902,
            },
            origin=None)

        with pytest.raises(EventFactory.BrokenRequest) as e:
            input.parse_body(request, command_name='MAKE_IT')

        assert e.value.event == 'BODY_DID_NOT_VALIDATE'
        assert e.value.data == {
            '@type': 'error',
            '@event': 'BODY_DID_NOT_VALIDATE',
            '@authorizer': {
                'user_id': 902,
            },
            'errors': {
                'amount': [
                    'Ensure this value is less than or equal to 19.'
                ]
            }
        }
        assert not e.value.is_critical

        # -- amount is title
        request = Mock(
            body=b'{"amount": 19}',
            origin=None,
            log_authorizer={
                'user_id': 902,
            })

        with pytest.raises(EventFactory.BrokenRequest) as e:
            input.parse_body(request, command_name='MAKE_IT')

        assert e.value.event == 'BODY_DID_NOT_VALIDATE'
        assert e.value.data == {
            '@type': 'error',
            '@event': 'BODY_DID_NOT_VALIDATE',
            '@authorizer': {
                'user_id': 902,
            },
            'errors': {
                'title': ['This field is required.']
            }
        }
        assert not e.value.is_critical




from unittest.mock import Mock, call

import search


def test_rank__to_sql(mocker):

    vector, query, compiler, connection = Mock(), Mock(), Mock(), Mock()
    as_sql = mocker.patch('search.rank.SearchRank.as_sql')

    rank = search.Rank(vector, query)
    result = rank.as_sql(compiler, connection)

    assert (
        as_sql.call_args_list ==
        [call(
            compiler,
            connection,
            function='ts_rank',
            template='%(function)s(%(expressions)s, 16 | 32)')])
    assert result == as_sql.return_value


def test_rank_cd__to_sql(mocker):

    vector, query, compiler, connection = Mock(), Mock(), Mock(), Mock()
    as_sql = mocker.patch('search.rank.SearchRank.as_sql')

    rank = search.RankCD(vector, query)
    result = rank.as_sql(compiler, connection)

    assert (
        as_sql.call_args_list ==
        [call(
            compiler,
            connection,
            function='ts_rank_cd',
            template='%(function)s(%(expressions)s, 4)')])
    assert result == as_sql.return_value




from unittest.mock import call

from django.test import TestCase
from click.testing import CliRunner
import pytest

from lily.base.utils import normalize_indentation
from lily.docs.management.commands.render_angular import command
from lily.docs.renderers.angular.renderer import AngularClientRenderer


class RenderAngularTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker):
        self.mocker = mocker

    def setUp(self):
        self.runner = CliRunner()

    def test_command__makes_the_right_calls(self):

        renderer = self.mocker.patch(  # noqa
            'lily.docs.management.commands.render_angular'
            '.AngularClientRenderer')

        result = self.runner.invoke(
            command, ['origin', 'super', '--only_build'])

        assert result.exit_code == 0
        assert normalize_indentation(
            result.output, 0
        ) == normalize_indentation(
            '''
                - Successfully rendered and built Angular Client [NO PUSHING TO REMOTE]
            ''', 0)  # noqa
        assert renderer.call_args_list == [call('origin', 'super')]
        assert renderer.return_value.render.call_args_list == [
            call(exclude_domains=(), include_domains=(), only_build=True),
        ]

    def test_command__default_values(self):

        render = self.mocker.patch.object(AngularClientRenderer, 'render')
        render.return_value = '1.0.19'

        result = self.runner.invoke(command, ['origin', 'super'])

        assert result.exit_code == 0
        assert normalize_indentation(
            result.output, 0
        ) == normalize_indentation(
            '''
                - Successfully rendered and pushed Angular Client version: 1.0.19
            ''', 0)  # noqa
        assert render.call_args_list == [
            call(exclude_domains=(), include_domains=(), only_build=None),
        ]

    def test_command__include_and_exclude_domains(self):

        render = self.mocker.patch.object(AngularClientRenderer, 'render')
        render.return_value = '1.0.19'

        result = self.runner.invoke(
            command,
            [
                'origin',
                'super',
                '--only_build',
                '--include_domain', 'CARDS',
                '--exclude_domain', 'PATHS',
                '--exclude_domain', 'GOSSIP',
            ])

        assert result.exit_code == 0
        assert render.call_args_list == [
            call(
                exclude_domains=('PATHS', 'GOSSIP'),
                include_domains=('CARDS',),
                only_build=True,
            ),
        ]




import json

from click.testing import CliRunner
from django.test import TestCase
import pytest

from lily.entrypoint.management.commands.render_commands import command
from lily.entrypoint.renderer import CommandsRenderer
from lily.entrypoint.serializers import CommandSerializer
from tests.factory import EntityFactory


ef = EntityFactory()


class RenderCommandsTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def initfixtures(self, mocker, tmpdir):
        self.mocker = mocker

        self.lily_dir = tmpdir.mkdir('.lily')

    def setUp(self):
        self.runner = CliRunner()

    def test_command__no_next_version__makes_the_right_calls(self):

        class ConfigMock:

            @classmethod
            def get_lily_path(cls):
                return str(self.lily_dir)

            @property
            def version(self):
                return '8.9.1'

            @property
            def next_version(self):
                return None

        self.mocker.patch(
            'lily.entrypoint.management.commands.render_commands.Config',
            ConfigMock)

        c0 = ef.command(domain_id='clients')
        c1 = ef.command(domain_id='products')

        self.mocker.patch.object(
            CommandsRenderer,
            'render'
        ).return_value = {
            '@enums': [{'enum_name': 'C'}],
            'BULK_READ_CLIENTS': c0,
            'DELETE_PRODUCT': c1,
        }

        result = self.runner.invoke(command)

        assert result.exit_code == 0
        commands_json = self.lily_dir.join('commands').join('8.9.1.json')
        assert (
            result.output.strip() ==
            f'Commands rendered for to file {str(commands_json)}')
        assert json.loads(commands_json.read()) == {
            '@enums': [{'enum_name': 'C'}],
            'BULK_READ_CLIENTS': CommandSerializer(c0).data,
            'DELETE_PRODUCT': CommandSerializer(c1).data,
        }

    def test_command__next_version__makes_the_right_calls(self):

        class ConfigMock:

            @classmethod
            def get_lily_path(cls):
                return str(self.lily_dir)

            @property
            def version(self):
                return '8.9.1'

            @property
            def next_version(self):
                return '8.9.2'

        self.mocker.patch(
            'lily.entrypoint.management.commands.render_commands.Config',
            ConfigMock)

        c0 = ef.command(domain_id='clients')
        c1 = ef.command(domain_id='products')

        self.mocker.patch.object(
            CommandsRenderer,
            'render'
        ).return_value = {
            '@enums': [],
            'BULK_READ_CLIENTS': c0,
            'DELETE_PRODUCT': c1,
        }

        result = self.runner.invoke(command)

        assert result.exit_code == 0
        commands_json = self.lily_dir.join('commands').join('8.9.2.json')
        assert (
            result.output.strip() ==
            f'Commands rendered for to file {str(commands_json)}')
        assert json.loads(commands_json.read()) == {
            '@enums': [],
            'BULK_READ_CLIENTS': CommandSerializer(c0).data,
            'DELETE_PRODUCT': CommandSerializer(c1).data,
        }




from unittest import TestCase

from lily.async import AsyncTask


class AsyncTaskTestCase(TestCase):

    def test_init(self):

        def fn():
            pass

        task = AsyncTask(callback=fn, args=[9, 1])

        assert task.callback == fn
        assert task.args == [9, 1]
        assert task.successful is False
        assert task.result is None




import json

from .events import EventFactory
from ..search import TextVector


class Input(EventFactory):

    class InputAttrs:
        pass

    def __init__(
            self,
            query_parser=None,
            body_parser=None):

        self.query_parser = query_parser
        self.body_parser = body_parser

    def __eq__(self, other):
        return (
            self.query_parser == other.query_parser and
            self.body_parser == other.body_parser)

    def parse(self, request, command_name):
        request.input = self.InputAttrs()

        # -- query parsing
        if self.query_parser:
            request.input.query = self.parse_query(request)

        # -- body parsing
        if self.body_parser:
            request.input.body, request.input.body_raw = self.parse_body(
                request, command_name=command_name)

    def parse_query(self, request):

        parsed = self.query_parser.as_query_parser(data=request.GET)

        if not parsed.is_valid():
            raise self.BrokenRequest(
                'QUERY_DID_NOT_VALIDATE',
                context=request,
                data={'errors': parsed.errors})

        else:
            return parsed.data

    def parse_body(self, request, command_name):
        try:
            body = str(request.body, encoding='utf8') or '{}'
            body = TextVector().transform_special_characters(body)

            data = json.loads(body)

        except (TypeError, ValueError):
            raise self.BrokenRequest(
                'BODY_JSON_DID_NOT_PARSE', context=request, is_critical=True)

        parsed = self.body_parser.as_body_parser(
            data=data,
            context={
                'request': request,
                'command_name': command_name,
            })
        if not parsed.is_valid():
            raise self.BrokenRequest(
                'BODY_DID_NOT_VALIDATE',
                context=request,
                data={'errors': parsed.errors})

        else:
            return parsed.data, data




import asyncio


from .task import AsyncTask


class BackoffExecutor:

    BACKOFF_UNIT = 3  # seconds

    BACKOFF_MAX_ATTEMPTS = 6

    def __init__(
            self,
            tasks: [AsyncTask],
            max_attempts=BACKOFF_MAX_ATTEMPTS,
            unit=BACKOFF_UNIT):

        self.tasks = tasks
        self.tasks_by_future = {}
        self.attempt_num = 0

        self.max_attempts = max_attempts
        self.unit = unit

    async def step(self, loop, attempt_num):

        futures = []
        idx = 0
        for task in self.tasks:
            if task.successful is False:
                futures.append(
                    loop.run_in_executor(
                        None,
                        task.callback,
                        *task.args
                    )
                )
                self.tasks_by_future[idx] = task
                idx += 1

        if futures:
            # -- sleep
            await asyncio.sleep(self.unit ** attempt_num - 1)

            future_results = asyncio.gather(*futures, return_exceptions=True)
            for i, result in enumerate(await future_results):
                task = self.tasks_by_future[i]
                if not isinstance(result, Exception):
                    task.successful = True

                task.result = result

        return (
            all([task.successful for task in self.tasks]),
            [task.result for task in self.tasks])

    def run(self):

        loop = asyncio.new_event_loop()

        for i in range(self.max_attempts):

            done, responses = loop.run_until_complete(self.step(loop, i))
            if done:
                loop.close()
                return responses

        loop.close()
        return responses




import glob
import importlib
import json
import os
import re

from lily_assistant.config import Config
import djclick as click


class Renderer:

    def render(self):

        dependencies = self.get_dependencies(
            self.get_most_current_migrations())

        return {
            'plan': self.get_migrations_plan(dependencies),
            'dependencies': {
                '.'.join(k): list(v)
                for k, v in dependencies.items()
            },
        }

    def get_most_current_migrations(self):
        base_path = os.path.join(os.getcwd(), Config().src_dir)

        filenames = glob.glob(os.path.join(base_path, '**/migrations/*.py'))
        filenames = [f for f in filenames if re.search(r'\d+\w+\.py$', f)]
        filenames = sorted(filenames, reverse=True)

        most_current_migrations = []
        apps_visited = set()
        for f in filenames:
            module = f
            module = re.sub(f'{os.getcwd()}/', '', module)
            module = re.sub(r'(.py$|\.\/)', '', module)
            module = re.sub(r'\/', '.', module)
            module_parts = module.split('.')

            app_name, migration = module_parts[-3], module_parts[-1]

            if app_name not in apps_visited:
                apps_visited.add(app_name)
                most_current_migrations.append((app_name, migration, module))

        return most_current_migrations

    def get_dependencies(self, most_current_migrations):

        return {
            (app_name, migration): set(
                importlib.import_module(module).Migration.dependencies)
            for app_name, migration, module in most_current_migrations
        }

    def get_migrations_plan(self, dependencies):

        all_migrations = sorted(dependencies.keys())
        migrations_plan = []
        while all_migrations:
            current_migration = all_migrations.pop(0)

            if current_migration not in migrations_plan:
                deps = dependencies.get(current_migration)

                # -- if has no dependencies OR
                # -- all of it's dependencies are already in the
                # -- `migrations_plan`
                sub_deps = deps & set(dependencies.keys())
                if not deps or sub_deps.issubset(set(migrations_plan)):
                    migrations_plan.append(current_migration)

                else:
                    all_migrations.append(current_migration)

        return migrations_plan


@click.command()
def command():

    config = Config()

    migrations_dir_path = os.path.join(config.get_lily_path(), 'migrations')
    if not os.path.exists(migrations_dir_path):
        os.mkdir(migrations_dir_path)

    version = config.next_version or config.version
    migrations_path = os.path.join(migrations_dir_path, f'{version}.json')
    with open(migrations_path, 'w') as f:
        f.write(json.dumps(Renderer().render(), indent=4, sort_keys=False))

    click.secho(
        f'Migrations plan rendered for to file {migrations_path}',
        fg='green')




from django.test import TestCase

from lily.base.access import Access, AccessSerializer


class AccessTestCase(TestCase):

    def test_defaults(self):

        a = Access()

        assert a.access_list is None
        assert a.is_private is False
        assert a.is_external is False

    def test_arguments_are_saved(self):

        a = Access(
            access_list=['SUPER_USER'],
            is_private=True,
            is_external=True)

        assert a.access_list == ['SUPER_USER']
        assert a.is_private is True
        assert a.is_external is True


class AccessSerializerTestCase(TestCase):

    def test_serialize(self):

        a = Access(
            access_list=['SUPER_USER'],
            is_private=True,
            is_external=True)

        assert AccessSerializer(a).data == {
            '@type': 'access',
            'is_private': True,
            'is_external': True,
            'access_list': ['SUPER_USER'],
        }




from . import serializers


class Access:
    def __init__(self, access_list=None, is_private=False, is_external=False):
        self.is_private = is_private
        self.access_list = access_list
        self.is_external = is_external

    def __eq__(self, other):
        return (
            self.is_private == other.is_private and
            self.access_list == other.access_list and
            self.is_external == other.is_external)


class AccessSerializer(serializers.Serializer):

    _type = 'access'

    access_list = serializers.ListField(child=serializers.CharField())

    is_private = serializers.BooleanField()

    is_external = serializers.BooleanField()




import importlib
import re
from time import time


def import_from_string(path):
    """Attempt to import a class from a string representation.

    From: django rest_framework

    """
    try:
        parts = path.split('.')
        module_path, class_name = '.'.join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)

    except ImportError as e:
        msg = "Could not import '{}' for setting. {}: {}.".format(
            path, e.__class__.__name__, e)
        raise ImportError(msg)


def normalize_indentation(text, min_indent=4):
    """Normalize text so that it's indented by specific amount.

    The minimal indent would be `min_indent`.

    """

    # -- trim beginning and ending
    text = re.sub(r'^\s*\n', '', text)
    text = re.sub(r'\n\s*$', '', text)

    # -- remove minimal indentation
    lines = text.split('\n')
    try:
        base_indent = min([
            len(re.match(r'\s*', line).group())
            for line in lines
            if line != ''])

    except ValueError:
        base_indent = 0

    pattern = re.compile('^\\s{{{}}}'.format(base_indent))
    lines = [pattern.sub('', line) for line in lines]

    # -- add `min_indent`
    lines = [min_indent * ' ' + line for line in lines]
    return '\n'.join(lines)


class Timer:

    def __init__(self):
        self.start = time()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    @property
    def elapsed(self):
        total_seconds = time() - self.start

        # -- hours
        hours = int(total_seconds // 3600)
        total_seconds = total_seconds - 3600 * hours

        # -- minutes
        minutes = int(total_seconds // 60)
        total_seconds = total_seconds - 60 * minutes

        # -- seconds
        seconds = int(total_seconds)
        total_seconds = total_seconds - int(total_seconds)

        # -- miliseconds
        miliseconds = int(round(1000 * total_seconds, 3))

        return '{hours}:{minutes}:{seconds}.{miliseconds}'.format(
            hours=str(hours).rjust(2, '0'),
            minutes=str(minutes).rjust(2, '0'),
            seconds=str(seconds).rjust(2, '0'),
            miliseconds=str(miliseconds).rjust(3, '0'))




from unittest.mock import Mock

from django.test import TestCase
import pytest

from lily.base.output import Output


class OutputTestCase(TestCase):

    def test_required_fields(self):
        # -- missing serializer
        with pytest.raises(TypeError):
            Output()

    def test_arguments_are_saved(self):
        serializer = Mock()

        o = Output(serializer=serializer)

        assert o.serializer == serializer




import json
import os

from lily_assistant.config import Config
from django.core import management
import djclick as click


@click.command()
@click.argument('v', type=str)
def command(v):

    config = Config()

    migrations_dir_path = os.path.join(config.get_lily_path(), 'migrations')

    migrations_path = os.path.join(migrations_dir_path, f'{v}.json')
    with open(migrations_path, 'r') as f:
        migrations_plan = json.loads(f.read())['plan']

    for app_name, migration in migrations_plan:
        management.call_command('migrate', app_name, migration)

    click.secho(
        f'Migrations plan for version {v} applied.',
        fg='green')




import djclick as click

from ...renderers.angular.renderer import AngularClientRenderer
from lily.base.utils import normalize_indentation


@click.command()
@click.argument(
    'client_origin',
    type=str)
@click.argument(
    'client_prefix',
    type=str)
@click.option(
    '--only_build',
    default=False,
    is_flag=True,
    type=click.BOOL,
    help='if true it only builds the client without pushing it to the remote')
@click.option(
    '--include_domain',
    '-i',
    type=str,
    multiple=True,
    help='provide domains to include')
@click.option(
    '--exclude_domain',
    '-e',
    type=str,
    multiple=True,
    help='provide domains to exclude')
def command(
        client_origin,
        client_prefix,
        only_build,
        include_domain,
        exclude_domain):
    """Render Angular Client based on the declared commands."""

    rendered_version = AngularClientRenderer(
        client_origin, client_prefix
    ).render(
        only_build=only_build,
        include_domains=include_domain,
        exclude_domains=exclude_domain)

    if only_build:
        click.secho(normalize_indentation('''
            - Successfully rendered and built Angular Client [NO PUSHING TO REMOTE]
        ''', 0), fg='green')  # noqa

    else:
        click.secho(normalize_indentation('''
            - Successfully rendered and pushed Angular Client version: {version}
        '''.format(version=rendered_version), 0), fg='green')  # noqa




from unittest.mock import Mock, call

import pytest

import search


@pytest.mark.django_db
@pytest.mark.parametrize("value, expected", [

    # case 0 - single word input - no transformation required
    ('hi', 'hi'),

    # case 1 - two words input should be separated by | operator
    ('hi all', 'all | hi'),

    # case 2 - many words input should be separated by | operator
    ('hi all how are you doing', 'all | are | doing | hi | how | you'),

    # case 3 - all punctuation characters are removed and used as split pivotal
    # points
    (
        'hi <>.all||&how&^%%are$@\' you !"|\\/ doing',
        'all | are | doing | hi | how | you'
    ),

    # case 4 - hashtags are not removed (hash character is replaced) and
    # presence of hashtags is treated as request for filtering
    # single hashtag
    ('#universe is crazy!', 'hhuniverse & (crazy | is)'),

    # case 5 - hashtags are not removed (hash character is replaced) and
    # presence of hashtags is treated as request for filtering -
    # many hashtags
    (
        '#universe is #crazy and #beautiful',
        'hhuniverse & hhcrazy & hhbeautiful & (and | is)',
    ),

    # case 6 - only hashtags
    (
        '#universe #crazy',
        'hhuniverse & hhcrazy',
    ),

], ids=list(map(str, range(0, 7))))
def test_query__parse_value(value, expected, mocker):
    mocker.patch('search.query.HASHTAG_ESCAPE_SEQUENCE', 'hh')

    q = search.Query('')

    assert expected == q.parse_value(value, 'simple')


@pytest.mark.django_db
@pytest.mark.parametrize("value, language_conf, expected", [

    # case 0 - with auto language detection and latex - pl
    (
        r'witam $$ 2 * x = \\pi $$',
        'polish',
        '2 | liczba | mnozone | mnozyc | mnożyć | pi | raz | rowna | '
        'rownac | rowny | równać | równy | witac | witać | x',
    ),

    # case 1 - with auto language detection and latex - en
    (
        r'hi there $$ 2 * x = \\pi $$',
        'english',
        '2 | equal | hi | number | pi | time | x',
    ),

    # case 2 - with auto language detection and latex - unsupported language
    # (for latex parsing) defaults to English
    (
        r'hi there $$ 2 * x = \\pi $$',
        'french',
        '2 | equal | hi | numb | pi | ther | tim | to | x',
    ),

], ids=['0', '1', '2'])
def test_query__get_transformed_value__with_latex(
        value, language_conf, expected, mocker):

    mocker.patch(
        'search.query.detector.detect_db_conf'
    ).return_value = language_conf

    q = search.Query('')

    assert expected == q.parse_value(value, language_conf)


@pytest.mark.django_db
def test_query__to_sql__with_config(mocker):
    mocker.patch(
        'search.query.detector.detect_db_conf'
    ).return_value = 'simple'

    q = search.Query('hi there')
    compiler = Mock(compile=Mock(return_value=['simple', ['a', 'b']]))
    connection = Mock()

    template, params = q.as_sql(compiler, connection)

    assert template == 'to_tsquery(simple::regconfig, %s)'
    assert params == ['a', 'b', 'hi | there']


@pytest.mark.django_db
def test_query__auto_detects_language(mocker):
    mocker.patch(
        'search.query.detector.detect_db_conf'
    ).return_value = 'polish'
    parse_value_mock = mocker.patch.object(
        search.Query, 'parse_value')
    parse_value_mock.return_value = 'witaj | świecie'
    compiler = Mock(compile=Mock(return_value=['polish', []]))

    q = search.Query('witaj świecie!')
    connection = Mock()

    template, params = q.as_sql(compiler, connection)

    assert template == 'to_tsquery(polish::regconfig, %s)'
    assert params == ['witaj | świecie']
    assert (
        parse_value_mock.call_args_list == [call('witaj świecie!', 'polish')])




import re


LATEX_PATTERN = re.compile(r'(\$\$(.+?)\$\$)')


MAX_ITERATIONS_COUNT = 20
"""
Maximum number of allowed iterations of transformations loop. Preventing
from looping too much over highly nested formulas.

"""

__PATTERN = r"""
    \s*(?P<%s>
    [0-9A-Za-z]|                           # single number or character
    \\[0-9A-Za-z]+((\s*\{.*\}\s*){1,2})?|  # any latex keyword or expression
                                           # with one or two arguments
                                           # enclosed within {}
    \{\s*.*?\s*\}                          # any expression enclosed within
                                           # { } braces
                                           # or already replaced string
    )%s\s*
"""

ARG_PATTERN = lambda n, o = False: __PATTERN % (n, (o and '?') or '')  # noqa
"""
Pattern of allowed values which can be found as argument of subscripts
operator ``_`` or superscript operator ``^``.

"""


EN_TRIGONOMETRY_FN_NAMES = {
    'sin': 'sine',
    'cos': 'cosine',
    'tan': 'tangent',
    'cot': 'cotangent',
}
PL_TRIGONOMETRY_FN_NAMES = {
    'sin': 'sinus',
    'cos': 'kosinus',
    'tan': 'tangens',
    'cot': 'kotangens',
}

PL = 'polish'
EN = 'english'
SUPPORTED_LANGUAGES = (EN, PL)


DEFAULT_LANGUAGE = 'english'


def decorate_transformations_repl(transformations):

    for transformation in transformations:
        replacements = transformation['replacements']

        for language, replacement in replacements.items():
            is_latex_function = transformation.get('is_latex_function', False)
            shortcut = transformation.get('shortcut', '')
            if isinstance(replacement, str):

                replacement = decorate_str_repl(
                    replacement, is_latex_function, shortcut)

            else:
                replacement = decorate_function_repl(
                    replacement, is_latex_function, shortcut)

            transformation['replacements'][language] = replacement

    return transformations


def decorate_function_repl(fn, is_latex_function, shortcut):

    def inner(m):
        return _decorate_repl(fn(m), is_latex_function, shortcut)

    return inner


def decorate_str_repl(s, is_latex_function, shortcut):
    return _decorate_repl(s, is_latex_function, shortcut)


def _decorate_repl(s, is_latex_function=False, shortcut=None):
    if is_latex_function:
        return r' \\REPL{%s %s} ' % (shortcut, s)
    else:
        return r' %s %s ' % (shortcut, s)


TRANSFORMATIONS = decorate_transformations_repl([

    #
    # Special symbols and modifiers
    #
    {
        'pattern': r'\\infty',
        'shortcut': 'infty',
        'replacements': {
            EN: r'infinity',
            PL: r'nieskończoność',
        },
    },
    {
        'pattern': r'\\dim',
        'shortcut': 'dim',
        'replacements': {
            EN: r'dimension of',
            PL: r'wymiar',
        },
    },
    {
        'pattern': r'\\degree',
        'shortcut': 'degree',
        'replacements': {
            EN: r'degrees',
            PL: r'stopni',
        },
    },
    {
        'pattern': r'\\to',
        'replacements': {
            EN: r'to ',
            PL: r'dążące do',
        },
    },
    {
        'pattern': r'\\limits',
        'replacements': {
            EN: r'in limits',
            PL: r'w granicach',
        },
    },

    #
    # Accent Functions
    #
    {
        'pattern': r'\\vec',
        'replacements': {
            EN: r'vector',
            PL: r'wektor',
        },
    },
    {
        'pattern': r'\\dot',
        'replacements': {
            EN: r'time derivative of',
            PL: r'pochodna po czasie z',
        },
    },
    {
        'pattern': r'\\bar',
        'replacements': {
            EN: r'average',
            PL: r'średnia',
        },
    },
    {
        'pattern': r'\\hat',
        'replacements': {
            EN: r'operator',
            PL: r'operator',
        },
    },


    #
    # Arithmetic
    #
    {
        'pattern': r'\+',
        'replacements': {
            EN: r'plus',
            PL: r'plus',
        },
    },
    {
        'pattern': r'\-',
        'replacements': {
            EN: r'minus',
            PL: r'minus',
        },
    },
    {
        'pattern': r'\*',
        'replacements': {
            EN: r'times',
            PL: r'razy mnożone przez',
        },
    },
    {
        'pattern': r'\\times',
        'replacements': {
            EN: r'times',
            PL: r'razy mnożone przez',
        },
    },
    {
        'pattern': r'\/',
        'replacements': {
            EN: r'divided by',
            PL: r'dzielone przez',
        }
    },
    {
        'pattern': r'\=',
        'replacements': {
            EN: r'equals to ',
            PL: r'równa się',
        },
    },
    {
        'pattern': r'\\neq',
        'replacements': {
            EN: r'not equals to ',
            PL: r'jest różne od',
        },
    },

    #
    # Number Relations
    #
    {
        'pattern': r'\\approx',
        'replacements': {
            EN: r'is approximately',
            PL: r'równa się w przybliżeniu',
        },
    },
    {
        'pattern': r'\\leq',
        'replacements': {
            EN: r'is less than equal to',
            PL: r'jest mniejsze lub równe',
        },
    },
    {
        'pattern': r'\\ll',
        'replacements': {
            EN: r'is much smaller than',
            PL: r'jest dużo mniejsze niż',
        },
    },
    {
        'pattern': r'\\geq',
        'replacements': {
            EN: r'is greater than equal to',
            PL: r'jest większe lub równe',
        },
    },
    {
        'pattern': r'\\gg',
        'replacements': {
            EN: r'is much greater than',
            PL: r'jest dużo większe niż',
        },
    },

    #
    # Fractions
    #
    {
        'pattern': r'\\[td]?frac{(.*)}{(.*)}',
        'is_latex_function': True,
        'replacements': {
            EN: r'fraction \1 divided by \2',
            PL: r'ułamek \1 dzielone przez \2',
        },
    },
    {
        'pattern': (
            r'\{' +
            ARG_PATTERN('num') +
            r'\\over' +
            ARG_PATTERN('den') +
            r'\}'),
        'is_latex_function': True,
        'replacements': {
            EN: r'fraction \g<num> divided by \g<den>',
            PL: r'ułamek \g<num> dzielone przez \g<den>',
        },
    },

    #
    # Complex Numbers
    #
    {
        'pattern': r'\\Im',
        'is_latex_function': True,
        'replacements': {
            EN: r'imaginary part of',
            PL: r'część urojona zespolona z',
        },
    },
    {
        'pattern': r'\\Re',
        'is_latex_function': True,
        'replacements': {
            EN: r'real part of',
            PL: r'część rzeczywista z',
        },
    },

    #
    # Exponentiation and Polynomials
    #

    # square
    {
        'pattern': r'\^\s*(2|\{\s*2\s*\})',
        'replacements': {
            EN: r'squared',
            PL: r'do kwadratu',
        },
    },

    # general numerical exponent
    {
        'pattern': (
            r'\^\s*('
            r'[0-13-9]+|'
            r'\{\s*[0-13-9]+\s*\}|'
            r'\{\s*(minus|plus|\+|\-)\s*[0-9]+\s*\}'
            r')'
        ),
        'replacements': {
            EN: r'to the power of \1',
            PL: r'podniesione do potęgi \1',
        },
    },

    # square root
    {
        'pattern': r'\\sqrt',
        'shortcut': 'sqrt',
        'is_latex_function': True,
        'replacements': {
            EN: r'square root function of',
            PL: r'funkcja pierwiastek kwadratowy z',
        },
    },

    #
    # Calculus
    #

    # limits
    {
        'pattern': r'\\lim(?!s|i)+',
        'shortcut': 'lim',
        'replacements': {
            EN: r'limit',
            PL: r'granica',
        },
    },
    {
        'pattern': r'\\liminf',
        'shortcut': 'liminf',
        'replacements': {
            EN: r'infimum limit',
            PL: r'granica infimum',
        },
    },
    {
        'pattern': r'\\limsup',
        'shortcut': 'limsup',
        'replacements': {
            EN: r'supremum limit',
            PL: r'granica supremum',
        },
    },
    {
        'pattern': r'\\inf',
        'shortcut': 'inf',
        'replacements': {
            EN: r'infimum of',
            PL: r'infimum z',
        },
    },
    {
        'pattern': r'\\sup',
        'shortcut': 'sup',
        'replacements': {
            EN: r'supremum of',
            PL: r'supremum z',
        },
    },
    {
        'pattern': r'\\max',
        'shortcut': 'max',
        'replacements': {
            EN: r'maximum of',
            PL: r'maksimum z',
        },
    },
    {
        'pattern': r'\\min',
        'shortcut': 'min',
        'replacements': {
            EN: r'minimum of',
            PL: r'minimum z',
        },
    },



    # sums
    {
        'pattern': r'\\sum',
        'shortcut': 'sum',
        'replacements': {
            EN: r'sum of',
            PL: r'suma',
        },
    },

    # products
    {
        'pattern': r'\\prod',
        'shortcut': 'prod',
        'replacements': {
            EN: r'product of',
            PL: r'iloczyn',
        },
    },

    # derivatives
    {
        'pattern': r'\\partial',
        'shortcut': 'partial',
        'replacements': {
            EN: r'partial derivative',
            PL: r'pochodna cząstkowa',
        },
    },

    # indefinite integral
    {
        'pattern': (
            r'\\((?P<o>o)?i{0,3}int(?!o)|smallint|intop)\s*'
            r'(?P<rest>[^_^]+)'),
        'shortcut': 'int',
        'replacements': {
            EN: (
                lambda m: '{o} indefinite integral of {rest}'.format(
                    o=(m.group('o') and 'countour') or '',
                    rest=m.group('rest'))),
            PL: (
                lambda m: 'całka nieoznaczona {o} z {rest}'.format(
                    o=(m.group('o') and 'po konturze') or '',
                    rest=m.group('rest'))),
        },
    },

    # definite integral
    # \int_A?^B pattern
    {
        'pattern': (
            r'\\((?P<o>o)?i{0,3}int(?!o)|smallint|intop)\s*' +
            r'\_' + ARG_PATTERN('sub') +
            r'\^?' + ARG_PATTERN('sup', True)),
        'shortcut': 'int',
        'is_latex_function': True,
        'replacements': {
            EN: (
                lambda m: (
                    r'{o} definite integral from {sub} to {sup} of'.format(
                        o=(m.group('o') and 'countour') or '',
                        sub=m.group('sub'),
                        sup=m.group('sup')))),
            PL: (
                lambda m: (
                    r'całka oznaczona {o} od {sub} do {sup} z'.format(
                        o=(m.group('o') and 'po konturze') or '',
                        sub=m.group('sub'),
                        sup=m.group('sup')))),
        },
    },

    # \int^B_A pattern
    {
        'pattern': (
            r'\\((?P<o>o)?i{0,3}int(?!o)|smallint|intop)\s*' +
            r'\^' + ARG_PATTERN('sup') +
            r'\_?' + ARG_PATTERN('sub', True)),
        'shortcut': 'int',
        'is_latex_function': True,
        'replacements': {
            EN: (
                lambda m: (
                    r'{o} definite integral from {sub} to {sup} of'.format(
                        o=(m.group('o') and 'countour') or '',
                        sub=m.group('sub'),
                        sup=m.group('sup')))),
            PL: (
                lambda m: (
                    r'całka oznaczona {o} od {sub} do {sup} z'.format(
                        o=(m.group('o') and 'po konturze') or '',
                        sub=m.group('sub'),
                        sup=m.group('sup')))),
        },
    },

    #
    # Greek letters
    #
    {
        'pattern': r'\\alpha',
        'is_latex_function': True,
        'replacements': {
            EN: r'alpha',
            PL: r'alfa',
        },
    },
    {
        'pattern': r'\\pi',
        'is_latex_function': True,
        'replacements': {
            EN: r'number pi',
            PL: r'liczba pi',
        },
    },
    {
        'pattern': r'\\Delta',
        'is_latex_function': True,
        'replacements': {
            EN: r'change delta of',
            PL: r'zmiana delta',
        },
    },
    {
        'pattern': r'\\Nabla',
        'replacements': {
            EN: r'divergence nabla of',
            PL: r'dywergencja nabla',
        },
    },

    #
    # Logarithms and Exponent functions
    #
    {
        'pattern': r'\\exp',
        'shortcut': 'exp',
        'replacements': {
            EN: 'exponent function of',
            PL: 'funkcja wykładnicza z',
        },
    },
    {
        'pattern': r'\\ln',
        'shortcut': 'ln',
        'replacements': {
            EN: 'natural logarithm function of',
            PL: 'funkcja logarytm naturalny z',
        },
    },
    {
        'pattern': r'\\log',
        'shortcut': 'log',
        'replacements': {
            EN: 'logarithm function of',
            PL: 'funkcja logarytm z',
        },
    },

    #
    # Trigonometry
    #
    # Since trigonometrical functions are using rather complex
    # pattern one inject their shortcut directly inside the repl
    {
        'pattern': r'\\(?P<arc>arc)?(?P<fn>sin|cos|cot|tan)(?P<h>h)?',
        'replacements': {
            EN: lambda m: r'{s} {arc}{h} {fn} function of'.format(
                arc=(m.group('arc') and 'inverse arc') or '',
                h=(m.group('h') and 'hyperbolic') or '',
                s=(
                    (m.group('arc') or '') +
                    m.group('fn') +
                    (m.group('h') or '')),
                fn=EN_TRIGONOMETRY_FN_NAMES.get(m.group('fn'), '')),
            PL: lambda m: r'{s} funkcja {arc} {fn} {h} z'.format(
                arc=(m.group('arc') and 'arkus') or '',
                h=(m.group('h') and 'hiperboliczny') or '',
                s=(
                    (m.group('arc') or '') +
                    m.group('fn') +
                    (m.group('h') or '')),
                fn=PL_TRIGONOMETRY_FN_NAMES.get(m.group('fn'), '')),
        },
    },

    #
    # Binomials
    #
    {
        'pattern': r'!',
        'replacements': {
            EN: r'factorial',
            PL: r'silnia',
        },
    },
    {
        'pattern': (
            r'\\[td]?binom' +
            ARG_PATTERN('up') +
            ARG_PATTERN('down')),
        'shortcut': 'binom',
        'is_latex_function': True,
        'replacements': {
            EN: r'binomial \g<up> choose \g<down>',
            PL: r'kombinacja \g<up> nad \g<down>'
        },
    },
    {
        'pattern': (
            r'\{' +
            ARG_PATTERN('up') +
            r'\\choose' +
            ARG_PATTERN('down') +
            r'\}'),
        'shortcut': 'choose',
        'is_latex_function': True,
        'replacements': {
            EN: r'binomial \g<up> choose \g<down>',
            PL: r'kombinacja \g<up> nad \g<down>',
        },
    },

    #
    # FUTURE: matrix notation
    #
    # Spec: should be able to parse simple matrix
    # Spec: should be able to parse matrix multiplication and other operations

    #
    # FUTURE: tensor notation
    #
    # Spec: Should be able to parse Einstein's general relativity equations
    # Spec: Should be able to parse Maxwell's equations written in tensor form
    # Spec: Should be able to parse Standard Model Lagrangian

    #
    # FUTURE: logic
    #
    # TODO: \forall \exists \nexists \therefore \because \equiv, \bigvee,
    # \bigwedge

])


def transform(text, language):
    """Transform provided text.

    Transform provided text by injecting after each occurrence of LaTeX
    syntax its human readable representation in a language of the text.

    """

    if language not in SUPPORTED_LANGUAGES:
        language = DEFAULT_LANGUAGE

    formulas = LATEX_PATTERN.findall(text)
    for raw_formula, formula_inner in set(formulas):
        text = text.replace(
            raw_formula,
            _transform_formula(formula_inner, language))

    return re.sub(r'\s+', ' ', text.strip())


def _transform_formula(formula, language):
    """Transform single LaTeX formula.

    Transform single LaTeX formula by iteratively replacing all occurrences
    of handled patterns and replacing them with human readable phrases.

    It has builtin maximum iteration mechanism which prevents parser
    from doing too many loops of replacement (loop which are meant to
    catch all nested formulas).

    """
    # initial cleaning
    formula = formula.replace('(', ' ').replace(')', ' ')

    iteration_no = 0
    current_transformations = TRANSFORMATIONS
    while current_transformations:
        next_transformations = []
        iteration_no += 1

        for transformation in current_transformations:

            formula, replacement_count = re.subn(
                transformation['pattern'],
                transformation['replacements'][language],
                formula.strip(),
                flags=re.VERBOSE)

            if replacement_count:
                next_transformations.append(transformation)

        current_transformations = next_transformations
        if iteration_no >= MAX_ITERATIONS_COUNT:
            break

    # final cleaning
    return re.sub(r'(REPL|\{|\}|\\|\,|\_|\^)', ' ', formula)




import logging
import json

from django.http import JsonResponse, HttpResponse


class JsonResponseBase(JsonResponse):

    status_code = NotImplemented

    data = {}

    def __init__(self, data={}):
        super(JsonResponseBase, self).__init__(data=(data or self.data))


class Json200(JsonResponseBase):
    status_code = 200


class Json201(JsonResponseBase):
    status_code = 201


class Json303(JsonResponseBase):
    status_code = 303


class Json400(JsonResponseBase):
    status_code = 400


class Json401(JsonResponseBase):
    status_code = 401


class Json403(JsonResponseBase):
    status_code = 403


class Json404(JsonResponseBase):
    status_code = 404


class Json409(JsonResponseBase):
    status_code = 409


class Json500(JsonResponseBase):
    status_code = 500


class HttpGenericResponse(HttpResponse):

    def __init__(self, status_code, content, *args, **kwargs):

        self.status_code = status_code
        super(HttpGenericResponse, self).__init__(content, *args, **kwargs)

        self._headers['content-type'] = ('content-type', 'application/json')


class EventFactory:

    def __init__(self):
        self.logger = logging.getLogger()

    class Context:

        def __init__(self, **kwargs):
            self.data = kwargs

        def is_empty(self):
            return not self.data

        def __eq__(self, other):
            return self.data == other.data

    class Generic(Exception):

        def __init__(self, status_code, content):
            self.status_code = status_code
            self.content = content
            self.logger = logging.getLogger()

        def extend(self, method, path):
            """Extend with extra attributes.

            Sometimes Success Exception is raised without all needed attributes
            since they're unknown at the instantiation time, but are know
            afterwards.

            This method enables one to load required attributes.

            """
            self.method = method
            self.path = path

            return self

        def log(self):
            # -- notify about the event
            message = '[{method} {path}] -> {status_code}'.format(
                method=self.method,
                path=self.path,
                status_code=self.status_code)
            self.logger.info(message)

            return self

        def response(self):
            return HttpGenericResponse(self.status_code, self.content)

    #
    # SUCCESS RESPONSES
    #
    class BaseSuccessException(Exception):

        verb = None

        def __init__(
                self,
                instance=None,
                data=None,
                event=None,
                output_context=None,
                context=None):

            self.context = context or EventFactory.Context()
            self.event = event
            self.instance = instance
            self.data = data or {}
            self.output_context = output_context or {}
            self.logger = logging.getLogger()

        def extend(self, event=None, context=None):
            """Extend with extra attributes.

            Sometimes Success Exception is raised without all needed attributes
            since they're unknown at the instantiation time, but are know
            afterwards.

            This method enables one to load required attributes.

            """

            if (self.context and
                    isinstance(self.context, EventFactory.Context) and
                    self.context.is_empty()):
                self.context = context

            if self.event is None:
                self.event = event

            return self

        def log(self):

            log_authorizer = getattr(self.context, 'log_authorizer', {})
            data = {
                '@event': self.event,
            }
            if log_authorizer:
                data['@authorizer'] = log_authorizer

            # -- notify about the event
            message = '{event}: {log_data}'.format(
                event=self.event,
                log_data=json.dumps(data))

            self.logger.info(message)

    class Executed(BaseSuccessException):
        response_class = Json200

    #
    # CRUD
    #
    class Created(BaseSuccessException):
        response_class = Json201

    class Read(BaseSuccessException):
        response_class = Json200

    class Updated(BaseSuccessException):
        response_class = Json200

    class Deleted(BaseSuccessException):
        response_class = Json200

    #
    # BULK CRUD
    #
    class BulkCreated(BaseSuccessException):

        verb = 'bulk_created'

        response_class = Json201

    class BulkRead(BaseSuccessException):

        verb = 'bulk_read'

        response_class = Json200

    class BulkUpdated(BaseSuccessException):

        verb = 'bulk_updated'

        response_class = Json200

    class BulkDeleted(BaseSuccessException):

        verb = 'bulk_deleted'

        response_class = Json200

    #
    # ERRORS RESPONSES
    #
    class BaseErrorException(Exception):

        extra_headers = None

        is_critical = None

        def __init__(
                self,
                event,
                context=None,
                data=None,
                is_critical=False):

            context = context or EventFactory.Context()
            self.event = event
            self.data = data or {}
            self.data.update({
                '@type': 'error',
                '@event': event,
            })

            log_authorizer = getattr(context, 'log_authorizer', {})
            if log_authorizer:
                self.data.update({
                    '@authorizer': log_authorizer,
                })

            self.logger = logging.getLogger()

            self.is_critical = is_critical

            # -- notify about the event
            message = '{event}: {data}'.format(
                event=event, data=json.dumps(data))

            # FIXME: !!!! make it log in the lazy way so that enriching with
            # the context could take place!!!
            if is_critical or self.is_critical:
                if hasattr(context, 'data'):
                    self.logger.error(
                        message, exc_info=True, extra={'data': context.data})

                else:
                    self.logger.error(
                        message, exc_info=True, extra={'request': context})

            else:
                # -- this will not trigger the sentry notification
                # -- since sentry logger is ignoring log with level below
                # -- ERROR
                self.logger.info(message)

        def update_with_context(self, context):
            log_authorizer = getattr(context, 'log_authorizer', {})
            if log_authorizer:
                self.data.update({
                    '@authorizer': log_authorizer,
                })

    class Redirect(BaseErrorException):
        response_class = Json303

        def __init__(self, *args, **kwargs):
            self.extra_headers = {
                'Location': kwargs.pop('redirect_uri'),
            }
            super(EventFactory.Redirect, self).__init__(*args, **kwargs)

    class BrokenRequest(BaseErrorException):
        response_class = Json400

    class AuthError(BaseErrorException):
        response_class = Json401

    class AccessDenied(BaseErrorException):
        response_class = Json403

    class DoesNotExist(BaseErrorException):
        response_class = Json404

    class Conflict(BaseErrorException):
        response_class = Json409

        is_critical = True

    class ServerError(BaseErrorException):
        response_class = Json500

        is_critical = True

    class Warning(BaseErrorException):
        """Log only Extra Exception.

        Is should serve only the logging purposes and therefore should never
        be used for the direct response creation.

        """

        response_class = NotImplementedError




from django.test import TestCase

from lily.docs.renderers.angular.signature import Signature


class MockInterface:

    def __init__(self, name=None, is_empty=False):
        self.name = name
        self._is_empty = is_empty

    def is_empty(self):
        return self._is_empty


class SignatureTestCase(TestCase):

    #
    # INPUT
    #
    def test_input__nothing(self):

        s = Signature(
            method='get',
            path='/cards',
            path_parameters=[],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True))

        assert s.input == ''

    def test_input__path_parameters(self):

        # -- single parameter
        s = Signature(
            method='get',
            path='/cards/{card_id}',
            path_parameters=[{'name': 'card_id', 'type': int}],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True))

        assert s.input == 'cardId: number'

        # -- multiple parameters
        s = Signature(
            method='get',
            path='/cards/{card_id}',
            path_parameters=[
                {'name': 'card_id', 'type': int},
                {'name': 'user_id', 'type': str},
                {'name': 'name', 'type': str},
            ],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True))

        assert s.input == 'cardId: number, userId: string, name: string'

    def test_input__query_params(self):

        s = Signature(
            method='get',
            path='/cards/',
            path_parameters=[],
            authorization_required=True,
            request_query=MockInterface(name='ReadCardRequestQuery'),
            request_body=MockInterface(is_empty=True))

        assert s.input == 'params: X.ReadCardRequestQuery'

    def test_input__path_parameters_and_query_params(self):

        s = Signature(
            method='get',
            path='/cards/{card_id}/users/{user_id}',
            path_parameters=[
                {'name': 'card_id', 'type': int},
                {'name': 'user_id', 'type': str},
            ],
            authorization_required=True,
            request_query=MockInterface(name='ReadCardRequestQuery'),
            request_body=MockInterface(is_empty=True))

        assert s.input == (
            'cardId: number, userId: string, params: X.ReadCardRequestQuery')

    def test_input__body_params(self):

        s = Signature(
            method='put',
            path='/paths/',
            path_parameters=[],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(name='ReadCardRequestBody'))

        assert s.input == 'body: X.ReadCardRequestBody'

    def test_input__path_parameters_and_body_params(self):

        s = Signature(
            method='put',
            path='/cards/{card_id}/users/{user_id}',
            path_parameters=[
                {'name': 'card_id', 'type': int},
                {'name': 'user_id', 'type': str},
            ],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(name='ReadCardRequestBody'))

        assert s.input == (
            'cardId: number, userId: string, body: X.ReadCardRequestBody')

    #
    # CALL_ARGS
    #
    def test_call_args__nothing(self):

        s = Signature(
            method='get',
            path='/cards',
            path_parameters=[],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True))

        assert s.call_args == "'/cards', { authorizationRequired: true }"

    def test_call_args__path_parameters(self):

        # -- single parameter
        s = Signature(
            method='get',
            path='/cards/{card_id}',
            path_parameters=[{'name': 'card_id', 'type': int}],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True))

        assert s.call_args == (
            '`/cards/${cardId}`, { authorizationRequired: true }')

        # -- multiple parameters
        s = Signature(
            method='get',
            path='/cards/{card_id}/{name}/users/{user_id}',
            path_parameters=[
                {'name': 'card_id', 'type': int},
                {'name': 'user_id', 'type': str},
                {'name': 'name', 'type': str},
            ],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True))

        assert s.call_args == (
            '`/cards/${cardId}/${name}/users/${userId}`, '
            '{ authorizationRequired: true }')

    def test_call_args__query_params(self):

        s = Signature(
            method='get',
            path='/paths/',
            path_parameters=[],
            authorization_required=True,
            request_query=MockInterface(name='ReadPathRequestQuery'),
            request_body=MockInterface(is_empty=True))

        assert s.call_args == (
            "'/paths/', { params, authorizationRequired: true }")

    def test_call_args__path_parameters_and_query_params(self):

        s = Signature(
            method='get',
            path='/paths/{path_id}/users/{user_id}',
            path_parameters=[
                {'name': 'path_id', 'type': int},
                {'name': 'user_id', 'type': str},
            ],
            authorization_required=True,
            request_query=MockInterface(name='ReadPathRequestQuery'),
            request_body=MockInterface(is_empty=True))

        assert s.call_args == (
            '`/paths/${pathId}/users/${userId}`, '
            '{ params, authorizationRequired: true }')

    def test_call_args__body_params(self):

        s = Signature(
            method='post',
            path='/cards/',
            path_parameters=[],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(name='ReadCardRequestBody'))

        assert s.call_args == (
            "'/cards/', body, { authorizationRequired: true }")

    def test_call_args__path_parameters_and_body_params(self):

        s = Signature(
            method='post',
            path='/cards/{card_id}/users/{user_id}',
            path_parameters=[
                {'name': 'card_id', 'type': int},
                {'name': 'user_id', 'type': str},
            ],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(name='ReadCardRequestBody'))

        assert s.call_args == (
            '`/cards/${cardId}/users/${userId}`, '
            'body, { authorizationRequired: true }')

    def test_call_args__path_parameters_and_no_body(self):

        s = Signature(
            method='post',
            path='/cards/{card_id}/users/{user_id}',
            path_parameters=[
                {'name': 'card_id', 'type': int},
                {'name': 'user_id', 'type': str},
            ],
            authorization_required=False,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True))

        assert s.call_args == (
            '`/cards/${cardId}/users/${userId}`, '
            '{}, { authorizationRequired: false }')

    def test_call_args__bulk_read_field_and_path_parameters(self):

        s = Signature(
            method='get',
            path='/cards/{card_id}/users/{user_id}',
            path_parameters=[
                {'name': 'card_id', 'type': int},
                {'name': 'user_id', 'type': str},
            ],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True),
            bulk_read_field='cards')

        assert (
            s.call_args ==
            '`/cards/${cardId}/users/${userId}`, '
            '{ authorizationRequired: true }')

    def test_call_args__bulk_read_field_and_path_parameters_and_query_params(self):  # noqa

        s = Signature(
            method='get',
            path='/cards/{card_id}',
            path_parameters=[
                {'name': 'card_id', 'type': int},
            ],
            authorization_required=True,
            request_query=MockInterface(is_empty=False),
            request_body=MockInterface(is_empty=True),
            bulk_read_field='cards')

        assert (
            s.call_args ==
            '`/cards/${cardId}`, { params, authorizationRequired: true }')

    #
    # CALL_ARGS_WITHOUT_PATH
    #
    def test_call_args_without_path__nothing(self):

        s = Signature(
            method='get',
            path='/cards',
            path_parameters=[],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True))

        assert s.call_args_without_path == ''

    def test_call_args_without_path__path_parameters(self):

        # -- single parameter
        s = Signature(
            method='get',
            path='/cards/{card_id}',
            path_parameters=[{'name': 'card_id', 'type': int}],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True))

        assert s.call_args_without_path == 'cardId'

        # -- multiple parameters
        s = Signature(
            method='get',
            path='/cards/{card_id}/{name}/users/{user_id}',
            path_parameters=[
                {'name': 'card_id', 'type': int},
                {'name': 'user_id', 'type': str},
                {'name': 'name', 'type': str},
            ],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(is_empty=True))

        assert s.call_args_without_path == 'cardId, userId, name'

    def test_call_args_without_path__query_params(self):

        s = Signature(
            method='get',
            path='/paths/',
            path_parameters=[],
            authorization_required=True,
            request_query=MockInterface(name='ReadPathRequestQuery'),
            request_body=MockInterface(is_empty=True))

        assert s.call_args_without_path == 'params'

    def test_call_args_without_path__path_parameters_and_query_params(self):

        s = Signature(
            method='get',
            path='/paths/{path_id}/users/{user_id}',
            path_parameters=[
                {'name': 'path_id', 'type': int},
                {'name': 'user_id', 'type': str},
            ],
            authorization_required=True,
            request_query=MockInterface(name='ReadPathRequestQuery'),
            request_body=MockInterface(is_empty=True))

        assert s.call_args_without_path == 'pathId, userId, params'

    def test_call_args_without_path__body_params(self):

        s = Signature(
            method='post',
            path='/cards/',
            path_parameters=[],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(name='ReadCardRequestBody'))

        assert s.call_args_without_path == 'body'

    def test_call_args_without_path__path_parameters_and_body_params(self):

        s = Signature(
            method='post',
            path='/cards/{card_id}/users/{user_id}',
            path_parameters=[
                {'name': 'card_id', 'type': int},
                {'name': 'user_id', 'type': str},
            ],
            authorization_required=True,
            request_query=MockInterface(is_empty=True),
            request_body=MockInterface(name='ReadCardRequestBody'))

        assert s.call_args_without_path == 'cardId, userId, body'



#!/usr/bin/env python

import json
import os.path
from setuptools import setup, find_packages


BASE_DIR = os.path.dirname(os.path.realpath(__file__))


#
# REQUIREMENTS
#
def parse_requirements(requirements):

    return [
        r.strip()
        for r in requirements
        if (
            not r.strip().startswith('#') and
            not r.strip().startswith('-e') and
            r.strip())
    ]


with open(os.path.join(BASE_DIR, 'requirements.txt')) as f:
    requirements = parse_requirements(f.readlines())


#
# CONFIG
#
with open(os.path.join(BASE_DIR, '.lily', 'config.json')) as f:
    config = json.loads(f.read())


setup(
    name=config['name'],
    packages=find_packages(),
    description='Lily MicroService Framework for Humans',
    url=config['repository'],
    version=config['version'],
    author='CoSphere Team',
    author_email='contact@cosphere.org',
    install_requires=requirements,
    data_files=[(
        '',
        [
            'requirements.txt',
            'README.md',
            '.lily/config.json',
            'lily/cli/base.makefile',
            'lily/search/detector/data/language_codes.yaml',
            'lily/search/detector/data/language_to_configuration_map.yaml',
            'lily/docs/renderers/markdown/base.md',
            'lily/search/stopwords/pl.json',
        ],
    )],
    include_package_data=True,
    entry_points='''
        [console_scripts]
        lily=lily.cli.cli:cli
    ''',
    keywords=['lily', 'microservice', 'django', 'drf'],
    classifiers=[])




from .query import Query  # noqa
from .rank import Rank, RankCD  # noqa
from .vector import (  # noqa
    StoredVector,
    OnTheFlyVector,
    TextVector,
    concatenate_tsvectors,
)
from .detector import detector  # noqa
from .latex.transformer import transform as transform_latex  # noqa
